from typing import Union, Dict, Any, List, Tuple, Optional
import pandas as pd
from tqdm.auto import tqdm
from .utils import divvy_workload, get_num_workers, process_queue
from pydantic import BaseModel
from multiprocessing import Pool
from copy import deepcopy
import json
from .vectors import VectorSequence, VectorMultiset
import pathlib


class LogIO(BaseModel):
    record_delimiter: Optional[str] = "[@@@]"
    left_token: Optional[str] = "[<<"
    key_value_delimiter: Optional[str] = "="
    right_token: Optional[str] = ">>]"
    parallelize_imports: Optional[bool] = True
    embedded_json_line_prefix: Optional[str] = "[[EMBEDDED_JSON_LINE]]"
    embedded_csv_line_prefix: Optional[str] = "[[EMBEDDED_CSV_LINE]]"  # spoiler alert
    log_formatter: Optional[
        str
    ] = "\n@@ {time:x} AT: {time} | LEVEL: {level} | IN: {name}.{function}\n\n{message} |\n"  # TODO: unify this object's string fields with Config?? # noqa

    def single_feature_to_log(self, key: str, value: str) -> str:
        return f"{self.left_token}{key}{self.key_value_delimiter}{value}{self.right_token}"

    def dict_to_log(self, key_value_dict: Dict[str, Any], include_delimiter: bool = False) -> str:
        s: str = self.record_delimiter if include_delimiter else ""
        for key, value in key_value_dict.items():
            s += f"{self.single_feature_to_log(key=key, value=value)}"
        return s


def parse_string_with_manual_tokens(
    input_string: str,
    tokens_dictionary: Dict[str, Tuple[str, str]],
    limit: int = None,
    record_delimiter: str = "[@@@]",
    disable_progress_bar: bool = False,
    parallelize_processing: Union[bool, int] = False,
) -> pd.DataFrame:
    """
    Extracts a pandas dataframe from a string

    :param input_string: string (e.g. logs file) to be parsed
    :param record_delimiter: The string that should be used to chunk up the string into observations / rows
    :param tokens_dictionary: Extraction rules. Key = label for column, value = (before token, after token)
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param limit: maximum number of rows to process
    :return: Vector data set extracted from the input string
    """

    # Initialize, split the string, and trim the list if longer than 'limit'
    row_buffers_list: List[Dict[str, Any]] = []
    raw_row_texts_list: List[str] = input_string.split(record_delimiter)
    if limit and (limit < len(raw_row_texts_list)):
        raw_row_texts_list = raw_row_texts_list[:limit]

    # Loop over rows
    p = tqdm(raw_row_texts_list, disable=disable_progress_bar)
    if not disable_progress_bar:
        p.set_description("Reading and parsing")

    for raw_row_text in p:
        row_buffer: Dict[str, Any] = dict()

        # Look for each key in this row
        for key, (before_token, after_token) in tokens_dictionary.items():
            if before_token in raw_row_text:
                target_substring = raw_row_text.split(before_token)[1]
                if after_token in target_substring:
                    row_buffer.setdefault(key, target_substring.split(after_token)[0])
        if row_buffer:
            row_buffers_list.append(row_buffer)

    # Convert the list of dictionaries to a dataframe (note: this is a slow step that I plan to optimize later)
    return list_of_dict_to_dataframe(
        data_point_dicts=row_buffers_list,
        disable_progress_bar=disable_progress_bar,
        parallelize_processing=parallelize_processing,
    )


def parse_file_with_manual_tokens(path: str, *args, **kwargs):
    """Wrapper for above function (parse_string_with_manual_tokens) that applies it to a file"""
    with open(path, "r") as f:
        return parse_string_with_manual_tokens(f.read(), *args, **kwargs)


def parse_string_with_key_value_delimiters(
    input_string: str,
    left_token: str = None,
    key_value_delimiter: str = None,
    right_token: str = None,
    record_delimiter: str = None,
    parallelize_read: Union[bool, int] = False,
    parallelize_processing: Union[bool, int] = False,
    limit: int = None,
    disable_progress_bar: bool = False,
) -> pd.DataFrame:
    """
    Extracts a data frame from a string

    Input: "the [@@@] quick [<<x='foo'>>] brown[<<y=9>>] [@@@]fox [<<y=93>>]"
    Output: (data frame)
                   x   y
            0  'foo'   9
            1    NaN  93

    Hint, write your logs using python fstrings like f"Foo [<<{x=}>>] and [<<{y=}>>]"

    :param input_string: string to be parsed
    :param record_delimiter: substring between rows
    :param left_token: left side of a record
    :param key_value_delimiter: marker between the variable name and its value
    :param right_token: right side of a record
    :param parallelize_read: whether to parallelize reading. Can be an integer (# workers) or bool True / False
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param limit: maximum number of rows to process
    :return: pandas.DataFrame or VectorMultiset or VectorSequence
    """

    # Use LogIO defaults for unspecified tokens
    if not left_token:
        left_token = LogIO().left_token
    if not right_token:
        right_token = LogIO().right_token
    if not key_value_delimiter:
        key_value_delimiter = LogIO().key_value_delimiter
    if not record_delimiter:
        record_delimiter = LogIO().record_delimiter

    # Initialize, split the string, and trim the list if longer than 'limit'
    row_buffers_list: List[Dict[str, Any]] = []
    raw_row_texts_list: List[str] = input_string.split(record_delimiter)
    if limit and (limit < len(raw_row_texts_list)):
        raw_row_texts_list = raw_row_texts_list[:limit]

    # Process the chunks
    num_workers: int = get_num_workers(parallelize_arg=parallelize_read)
    if parallelize_read and (num_workers > 1):
        batches: List[List[Any]] = divvy_workload(num_workers=num_workers, tasks=raw_row_texts_list)
        i: List[Tuple[List[str], str, str, str]] = [
            (b, left_token, key_value_delimiter, right_token) for b in batches
        ]
        with Pool(num_workers) as pool:
            chunk_buffers_nested: List[List[Dict[str, Any]]] = pool.map(
                func=multi_key_value_extraction_lambda, iterable=i
            )
        row_buffers_list: List[Dict[str, Any]] = [
            item for sublist in chunk_buffers_nested for item in sublist
        ]

    else:
        # Serial processing
        # Loop over rows
        p1 = tqdm(raw_row_texts_list, disable=disable_progress_bar)
        if not disable_progress_bar:
            p1.set_description("Scanning text")

        for chunk in p1:
            row_buffers_list.append(
                key_value_extraction_lambda(
                    chunk,
                    left_token=left_token,
                    right_token=right_token,
                    key_value_delimiter=key_value_delimiter,
                )
            )

    return list_of_dict_to_dataframe(
        row_buffers_list,
        disable_progress_bar=disable_progress_bar,
        parallelize_processing=parallelize_processing,
    )


def parse_file_with_key_value_delimiters(path: str, *args, **kwargs):
    """Wrapper for above function (parse_string_with_key_value_delimiters) that applies it to a file"""
    with open(path, "r") as f:
        return parse_string_with_key_value_delimiters(f.read(), *args, **kwargs)


def parse_string_with_embedded_json_unprocessed_dicts(
    input_string: str,
    embedded_json_line_prefix: str = None,
    limit: int = None,
    ignore_decode_errors: bool = False,
    end_of_line: str = "\n",
) -> List[Dict[Any, Any]]:
    """
    Extracts embedded json from log files

    :param input_string: string (e.g. logs file) to be parsed
    :param embedded_json_line_prefix: marker string at the start of every embedded json line
    :param ignore_decode_errors: if True, silently ignores rows that don't parse (e.g. if pulling from incomplete row)
    :param end_of_line: substring between rows
    :param limit: maximum number of rows to process
    """

    # Use LogIO defaults for unspecified tokens
    if not embedded_json_line_prefix:
        embedded_json_line_prefix = LogIO().embedded_json_line_prefix

    # Initialize, split the string, and trim the list if longer than 'limit'
    if embedded_json_line_prefix not in input_string:
        return []

    # Pull out and clean the embedded data
    embedded_with_trailing_data = input_string.split(embedded_json_line_prefix)[1:]
    if limit and (limit < len(embedded_with_trailing_data)):
        embedded_with_trailing_data = embedded_with_trailing_data[:limit]
    embedded_clean = [x.split(end_of_line)[0] if end_of_line in x else x for x in embedded_with_trailing_data]

    return_list: List[Dict[Any, Any]] = []
    for s in embedded_clean:
        try:
            return_list.append(json.loads(s))
        except json.JSONDecodeError as e:
            if not ignore_decode_errors:
                raise e
    return return_list


def parse_string_with_embedded_json(
    input_string: str,
    embedded_json_line_prefix: str = None,
    parallelize_processing: Union[bool, int] = False,
    disable_progress_bar: bool = False,
    limit: int = None,
    end_of_line: str = "\n",
) -> pd.DataFrame:
    """
    Extracts embedded json from log files

    :param input_string: string (e.g. logs file) to be parsed
    :param embedded_json_line_prefix: marker string at the start of every embedded json line
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param end_of_line: substring between rows
    :param limit: maximum number of rows to process
    """
    extracted_json_dicts: List[Dict[str, Any]] = parse_string_with_embedded_json_unprocessed_dicts(
        input_string=input_string,
        embedded_json_line_prefix=embedded_json_line_prefix,
        limit=limit,
        end_of_line=end_of_line,
    )
    return list_of_dict_to_dataframe(
        extracted_json_dicts,
        disable_progress_bar=disable_progress_bar,
        parallelize_processing=parallelize_processing,
    )


def parse_file_with_embedded_json(path: str, *args, **kwargs) -> pd.DataFrame:
    """Wrapper for above (parse_string_with_embedded_json) function that applies it to a file"""
    with open(path, "r") as f:
        return parse_string_with_embedded_json(f.read(), *args, **kwargs)


def key_value_extraction_lambda(
    input_chunk: str,
    left_token: str = None,
    key_value_delimiter: str = None,
    right_token: str = None,
) -> Dict[str, Any]:
    """Helper lambda so we can do this in parallel or series"""
    output_chunk_dict: Dict[str, Any] = dict()
    raw_breaks: List[str] = input_chunk.split(left_token)[1:]
    middle: List[str] = [x.split(right_token)[0] for x in raw_breaks]
    # Loop over possible tokens
    for entry in middle:
        if key_value_delimiter not in entry:
            continue  # Do not attempt to parse if the delimiter is not present
        substrings: List[str] = entry.split(key_value_delimiter)
        key: str = substrings[0]
        if not key:
            continue  # Do not continue to parse if the key is known
        value: str = "".join([x + key_value_delimiter for x in substrings[1:]])[:-1]
        output_chunk_dict.setdefault(key, value)  # Add the value to this row
    return output_chunk_dict


def multi_key_value_extraction_lambda(args: Tuple[List[str], str, str, str]) -> List[Dict[str, Any]]:
    """Helper function that works through a queue of chunks (for parallelization)"""
    input_chunks, left_token, key_value_delimiter, right_token = args
    return [
        key_value_extraction_lambda(
            x,
            left_token=left_token,
            key_value_delimiter=key_value_delimiter,
            right_token=right_token,
        )
        for x in input_chunks
    ]


def data_frame_init_wrapper(data: Any) -> pd.DataFrame:
    """
    Temp thin helper function to wrap pandas init (this only exists to be handy for the next function)

    :param data: data for the dataframe, e.g. a list of dictionaries
    :return: the pandas dataframe
    """
    return pd.DataFrame(data, index=range(len(data)))


def list_of_dict_to_dataframe(
    data_point_dicts,
    parallelize_processing: Union[bool, int] = True,
    disable_progress_bar: bool = False,
    pandas_automatic: bool = True,
    suppress_deprecation_warning: bool = False,
) -> pd.DataFrame:
    """
     Extracts a data frame from a string

    Input: "the [@@@] quick [<<x='foo'>>] brown[<<y=9>>] [@@@]fox [<<y=93>>]"
    Output: (data frame)
                   x   y
            0  'foo'   9
            1    NaN  93

    Hint, write your logs using python fstrings like f"Foo [<<{x=}>>] and [<<{y=}>>]"

    :param data_point_dicts: list of dictionaries to be added to a dataframe
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param pandas_automatic: Throw everything into pan
    :param suppress_deprecation_warning: Nudge users to adopt more efficient method, pass "True" to silence
    :return: pandas.DataFrame or VectorMultiset or VectorSequence
    """
    num_reshape_workers: int = get_num_workers(parallelize_arg=parallelize_processing)

    if pandas_automatic:
        if not parallelize_processing:
            return pd.DataFrame(data_point_dicts)
        if parallelize_processing:
            if not disable_progress_bar:
                print(
                    f"Reshaping data (step 1 of 2): processing in parallel with {num_reshape_workers} workers"
                )
            batches_of_dictionaries: List[List[Any]] = divvy_workload(
                num_workers=num_reshape_workers, tasks=data_point_dicts
            )
            list_of_dataframes: List[pd.DataFrame] = process_queue(
                data_frame_init_wrapper,
                batches_of_dictionaries,
                pool_function="map",
                batching=False,  # here, we batch manually above rather than inside multiprocess
                num_workers=num_reshape_workers,
            )
            if not disable_progress_bar:
                print("Reshaping data (step 2 of 2): attaching subframes")
            return pd.concat(list_of_dataframes, ignore_index=True)

    if not suppress_deprecation_warning:
        print(f"You are using a very slow method for reshaping the data")
        print(f"Recommendation: use list_of_dict_to_dataframe(..., pandas_automatic=True)")
        print(
            f"To suppress this warning, pass: use list_of_dict_to_dataframe(..., suppress_deprecation_warning=True)"
        )

    if parallelize_processing and (num_reshape_workers > 1):
        batches: List[List[Any]] = divvy_workload(num_workers=num_reshape_workers, tasks=data_point_dicts)
        if not disable_progress_bar:
            print(f"Reshaping data: processing in parallel with {num_reshape_workers} workers")
        with Pool(num_reshape_workers) as pool:
            dataframes: List[pd.DataFrame] = pool.map(func=dicts_to_dataframe, iterable=batches)
        df: pd.DataFrame = pd.concat(dataframes, ignore_index=True)
    else:
        df: pd.DataFrame = pd.DataFrame()
        # THIS LOOP IS SO SLOW - you should use pandas_automatic=True
        for chunk_buffer in (p2 := tqdm(data_point_dicts, disable=disable_progress_bar)):
            p2.set_description("Reshaping data:")
            if chunk_buffer:
                df: pd.DataFrame = pd.concat([df, pd.DataFrame(chunk_buffer, index=[-1])], ignore_index=True)
    return df


def dicts_to_dataframe(dictionaries: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    NB: THIS IS AN INEFFICIENT FUNCTION THAT SCALES AWFULLY, PLEASE USE list_of_dict_to_dataframe instead
    Helper function that converts a list of dictionaries into a dataframe (each dictionary = 1 row)

    :param dictionaries: list of dictionaries, with one value per key per row
    :return: dataframe representation
    """
    df: pd.DataFrame = pd.DataFrame()
    for dictionary in [d for d in dictionaries if d]:
        df = pd.concat([df, pd.DataFrame(dictionary, index=[-1])], ignore_index=True)
    return df


def batch_dicts_to_dataframe(dictionaries: List[Dict[str, Any]], batch_size: int = 1000) -> pd.DataFrame:
    """
    NB: THIS IS AN INEFFICIENT FUNCTION THAT SCALES AWFULLY, PLEASE USE list_of_dict_to_dataframe instead
    Helper function that converts a list of dictionaries into a dataframe (each dictionary = 1 row)

    :param dictionaries: list of dictionaries, with one value per key per row
    :param batch_size: how many rows to batch together
    :return: dataframe representation
    """

    batches: List[pd.DataFrame] = []
    df_in_progress: pd.DataFrame = pd.DataFrame()
    for i, dictionary in enumerate([d for d in dictionaries if d]):
        df_in_progress = pd.concat([df_in_progress, pd.DataFrame(dictionary, index=[-1])], ignore_index=True)
        if i % batch_size == 0:
            batches.append(deepcopy(df_in_progress))
            df_in_progress = pd.DataFrame()

    return pd.concat(batches, ignore_index=True)


#####################################
# EVERYTHING BELOW THIS LINE IS
# LEGACY AND SHOULD BE DEPRECATED
# (PLEASE INSTEAD USE ABOVE FUNCTIONS)
#####################################


# LEGACY FUNCTION - New use of this function is not recommended, however it works just fine
def auto_extract_from_text(
    input_string: str,
    return_type: str = "dataframe",
    left_token: str = None,
    key_value_delimiter: str = None,
    right_token: str = None,
    basis_col_name: str = None,
    record_delimiter: str = None,
    parallelize_read: Union[bool, int] = False,
    parallelize_processing: Union[bool, int] = False,
    limit: int = None,
    disable_progress_bar: bool = False,
) -> Union[pd.DataFrame, VectorSequence, VectorMultiset]:
    """
    Extracts a data frame from a string

    Input: "the [@@@] quick [<<x='foo'>>] brown[<<y=9>>] [@@@]fox [<<y=93>>]"
    Output: (data frame)
                   x   y
            0  'foo'   9
            1    NaN  93

    Hint, write your logs using python fstrings like f"Foo [<<{x=}>>] and [<<{y=}>>]"

    :param input_string: string to be parsed
    :param record_delimiter: substring between rows
    :param left_token: left side of a record
    :param key_value_delimiter: marker between the variable name and its value
    :param right_token: right side of a record
    :param basis_col_name: optional, specify a basis column name for a VectorSequence (otherwise ignored)
    :param return_type: what format to return the data in
    :param parallelize_read: whether to parallelize reading. Can be an integer (# workers) or bool True / False
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param limit: maximum number of rows to process
    :return: pandas.DataFrame or VectorMultiset or VectorSequence
    """

    if not left_token:
        left_token = LogIO().left_token
    if not right_token:
        right_token = LogIO().right_token
    if not key_value_delimiter:
        key_value_delimiter = LogIO().key_value_delimiter
    if not record_delimiter:
        record_delimiter = LogIO().record_delimiter

    record_chunks: List[str] = input_string.split(record_delimiter)[1:]
    if limit and (limit < len(record_chunks)):
        record_chunks = record_chunks[:limit]
    chunk_buffers: List[Dict[str, Any]] = []

    # Process the chunks
    num_workers: int = get_num_workers(parallelize_arg=parallelize_read)
    if parallelize_read and (num_workers > 1):
        batches: List[List[Any]] = divvy_workload(num_workers=num_workers, tasks=record_chunks)
        i: List[Tuple[List[str], str, str, str]] = [
            (b, left_token, key_value_delimiter, right_token) for b in batches
        ]
        with Pool(num_workers) as pool:
            chunk_buffers_nested: List[List[Dict[str, Any]]] = pool.map(
                func=multi_key_value_extraction_lambda, iterable=i
            )
        chunk_buffers: List[Dict[str, Any]] = [item for sublist in chunk_buffers_nested for item in sublist]

    else:
        # Serial processing
        for chunk in (p1 := tqdm(record_chunks, disable=disable_progress_bar)):
            p1.set_description("Scanning file (step 1 of 2)")
            chunk_buffers.append(
                key_value_extraction_lambda(
                    chunk,
                    left_token=left_token,
                    right_token=right_token,
                    key_value_delimiter=key_value_delimiter,
                )
            )

    df: pd.DataFrame = list_of_dict_to_dataframe(
        chunk_buffers,
        disable_progress_bar=disable_progress_bar,
        parallelize_processing=parallelize_processing,
    )

    if "dataframe" in return_type.lower():
        return df
    elif "multiset" in return_type.lower():
        return VectorMultiset(data=df)
    elif "sequence" in return_type.lower():
        return VectorSequence(data=df, basis_col_name=basis_col_name)


# LEGACY FUNCTION - New use of this function is not recommended, however it works just fine
def auto_extract_from_file(
    file_path: Union[str, pathlib.Path],
    record_delimiter: str = None,
    right_token: str = None,
    left_token: str = None,
    key_value_delimiter: str = None,
    return_type: str = "dataframe",
    basis_col_name: str = None,
    disable_progress_bar: bool = False,
    parallelize_read: Union[bool, int] = False,
    parallelize_processing: Union[bool, int] = False,
    **kwargs,
) -> Union[pd.DataFrame, VectorSequence, VectorMultiset]:
    """
    Extracts a data frame from a file

    Input: "the [@@@] quick [<<x='foo'>>] brown[<<y=9>>] [@@@]fox [<<y=93>>]"
    Output: (data frame)
                   x   y
            0  'foo'   9
            1    NaN  93

    Hint, write your logs using python fstrings like f"Foo [<<{x=}>>] and [<<{y=}>>]"

    :param file_path: file to read
    :param record_delimiter: substring between rows
    :param left_token: left side of a record
    :param key_value_delimiter: marker between name and
    :param right_token: right side of a record
    :param basis_col_name: optional, specify a basis column name for a VectorSequence (otherwise ignored)
    :param return_type: what format to return the data in
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param parallelize_read: whether to parallelize reading. Can be an integer (# workers) or bool True / False
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :return: pandas.DataFrame or VectorMultiset or VectorSequence
    """

    if not left_token:
        left_token = LogIO().left_token
    if not right_token:
        right_token = LogIO().right_token
    if not key_value_delimiter:
        key_value_delimiter = LogIO().key_value_delimiter
    if not record_delimiter:
        record_delimiter = LogIO().record_delimiter

    # Read in the file
    with open(file_path, "r") as f:
        input_string: str = f.read()

    return auto_extract_from_text(
        input_string=input_string,
        return_type=return_type,
        record_delimiter=record_delimiter,
        left_token=left_token,
        parallelize_read=parallelize_read,
        parallelize_processing=parallelize_processing,
        key_value_delimiter=key_value_delimiter,
        right_token=right_token,
        basis_col_name=basis_col_name,
        disable_progress_bar=disable_progress_bar,
        **kwargs,
    )


# LEGACY FUNCTION - New use of this function is not recommended, however it works just fine
def extract_text_to_dataframe(
    input_string: str,
    tokens_dictionary: Dict[str, Tuple[str, str]],
    limit: int = None,
    record_delimiter: str = "[@@@]",
    disable_progress_bar: bool = False,
    parallelize_processing: Union[bool, int] = False,
) -> pd.DataFrame:
    """
    Extracts a pandas dataframe from a string

    :param input_string: string (e.g. logs file) to be parsed
    :param record_delimiter: The string that should be used to chunk up the string into observations / rows
    :param tokens_dictionary: Extraction rules. Key = label for column, value = (before token, after token)
    :param disable_progress_bar: pass anything True to silence the progress bar
    :param parallelize_processing: whether to parallelize flattening. Can be an integer (# workers) or bool True / False
    :param limit: maximum number of rows to process
    :return: Vector data set extracted from the input string
    """
    # df_output: pd.DataFrame = pd.DataFrame()
    chunk_buffers_list: List[Dict[str, Any]] = []
    record_chunks: List[str] = input_string.split(record_delimiter)
    if limit and (limit < len(record_chunks)):
        record_chunks = record_chunks[:limit]
    for chunk in (p := tqdm(record_chunks, disable=disable_progress_bar)):
        p.set_description("Reading and parsing (step 1 of 2)")
        chunk_buffer: Dict[str, Any] = dict()
        for key, (before_token, after_token) in tokens_dictionary.items():
            if before_token in chunk:
                target_chunk = chunk.split(before_token)[1]
                if after_token in target_chunk:
                    chunk_buffer.setdefault(key, target_chunk.split(after_token)[0])
        if chunk_buffer:
            chunk_buffers_list.append(chunk_buffer)

    return list_of_dict_to_dataframe(
        data_point_dicts=chunk_buffers_list,
        disable_progress_bar=disable_progress_bar,
        parallelize_processing=parallelize_processing,
    )


# LEGACY FUNCTION - New use of this function is not recommended, however it works just fine
def extract_text_to_vector(
    input_string: str,
    tokens_dictionary: Dict[str, Tuple[str, str]],
    record_delimiter: str = "[@@@]",
    disable_progress_bar: bool = False,
    basis_col_name: str = None,
    **kwargs,
) -> Union[VectorMultiset, VectorSequence]:
    """
    Extracts a VectorMultiset from a string (or a VectorSequence if you specify `basis_col_name`)

    :param input_string: string (e.g. logs file) to be parsed
    :param record_delimiter: The string that should be used to chunk up the string into observations / rows
    :param tokens_dictionary: Extraction rules. Key = label for column, value = (before token, after token)
    :param basis_col_name: Optional - if specified returns a VectorSequence
    :param disable_progress_bar: pass anything True to silence the progress bar
    :return: Vector data set extracted from the input string
    """
    df_output: pd.DataFrame = extract_text_to_dataframe(
        input_string=input_string,
        record_delimiter=record_delimiter,
        tokens_dictionary=tokens_dictionary,
        disable_progress_bar=disable_progress_bar,
        **kwargs,
    )
    if basis_col_name:
        return VectorSequence(data=df_output, basis_col_name=basis_col_name)
    else:
        return VectorMultiset(data=df_output)


# LEGACY FUNCTION - New use of this function is not recommended, however it works just fine
def extract_file_to_vector(
    file_path: Union[str, pathlib.Path],
    record_delimiter: str,
    tokens_dictionary: Dict[str, Tuple[str, str]],
    disable_progress_bar: bool = False,
    basis_col_name: str = None,
    **kwargs,
) -> Union[VectorSequence, VectorMultiset]:
    """
    Wrapper for extract_from_text that pulls vector data out of a file, such as raw log output.

    :param file_path: file to read
    :param record_delimiter: The string that should be used to chunk up the string into observations / rows
    :param tokens_dictionary: Extraction rules. Key = label for column, value = (before token, after token)
    :param basis_col_name: Optional - if specified returns a VectorSequence
    :param disable_progress_bar: pass anything True to silence the progress bar
    :return: Vector data set extracted from the input string
    """
    with open(file_path, "r") as f:
        return extract_text_to_vector(
            input_string=f.read(),
            record_delimiter=record_delimiter,
            tokens_dictionary=tokens_dictionary,
            basis_col_name=basis_col_name,
            disable_progress_bar=disable_progress_bar,
            **kwargs,
        )



import datetime
from typing import List, Any, Union, Tuple, Callable, Dict

import matplotlib
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from mpl_toolkits.mplot3d import Axes3D
from itertools import cycle


from .config import Style
from .utils import (
    looks_like_list_of_lists,
    margin_calc,
    to_list_if_other_array,
    make_dict,
    dict_pretty,
)


##################
# Helper functions
##################


def plot_best_fit_line(
    x_data: Any,
    y_data: Any,
    degree: int = None,
    color: str = "k",
    style: Style = None,
    **kwargs,
) -> None:
    """Adds the line of best fit to a plot

    :param x_data: can be an array-like object or a list of array-like objects (for multiple traces)
    :param y_data: can be an array-like object or a list of array-like objects (for multiple traces)
    :param degree: degree oft the polynomial to fit
    :param color: color for the line
    :param style: optional style object (which contains a line width parameter)
    :param kwargs: additional keyword arguments for matplotlib.pyplot.plt
    :return: None
    """
    if kwargs.get("linewidth"):
        line_width = kwargs["linewidth"]
    else:
        if (not style) or (not style.linewidth):
            line_width: float = Style().linewidth
        else:
            line_width: float = style.linewidth
    if not degree:
        degree: int = 1
    z = np.polyfit(np.array(x_data), np.array(y_data), degree)
    plt.plot(
        x_data,
        [z[0] * x + z[1] for x in x_data],
        color=color,
        linewidth=line_width,
        **kwargs,
    )


def adjust_axes(
    log_axes: Union[str, List[str]] = "",
    style: Style = None,
    xlim: Any = None,
    ylim: Any = None,
    x_axis_human_tick_labels: bool = False,
    x_axis_formatter: str = None,
) -> None:
    """Helper function that adjusts the axes of a plot as specified

    :param log_axes: can look like: 'x' or 'y' or 'xy' // or look like: ['x'] or ['y'] or ['x', 'y']
    :param style: the Style object to apply
    :param xlim: bound for the x-axis
    :param ylim: bound for the y-axis
    :param x_axis_human_tick_labels: True if the x-axis labels should be displayed as human-readable
    :param x_axis_formatter: format string for the x-axis if human-readable
    :return: None
    """
    if not style:
        style = Style()
    if style.grid:
        plt.grid()
    if style.tight_axes:
        plt.autoscale(enable=True, axis="x", tight=True)
    if xlim:
        plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)
    if "x" in log_axes:
        plt.xscale("log")
    if "y" in log_axes:
        plt.yscale("log")
    if x_axis_human_tick_labels:
        if x_axis_formatter is None:
            x_axis_formatter: str = "%Y-%m-%d"
        ax = plt.gca()
        ax.xaxis.set_major_formatter(mdates.DateFormatter(x_axis_formatter))
        for tick_label in ax.get_xticklabels(which="major"):
            tick_label.set(rotation=30, horizontalalignment="right")


def apply_plot_labels(xlabel: str = "", ylabel: str = "", title: str = "", style: Style = None) -> None:
    """Helper function to apply labels (xlabel, ylabel, title) according to the style guide (including translation)

    :param xlabel: text or rosetta key for the x-axis label
    :param ylabel: text or rosetta key for the y-axis label
    :param title: text or rosetta key for the title
    :param style: config Style object
    """
    if not style:
        style: Style = Style()
    plt.xlabel(
        style.rosetta.translate(xlabel, missing_response="return_input"),
        fontsize=style.label_fontsize,
    )
    plt.ylabel(
        style.rosetta.translate(ylabel, missing_response="return_input"),
        fontsize=style.label_fontsize,
    )
    plt.title(
        style.rosetta.translate(title, missing_response="return_input"),
        fontsize=style.title_fontsize,
    )


def apply_watermark(watermark_text: str, style: Style = None, use_default: bool = True, **kwargs) -> None:
    """Helper function to apply watermark text to a plot based on config Style object parameters

    :param watermark_text: Text to display
    :param style: configuration object (Style)
    :type use_default: if True, then the config watermark_text will override the input argument watermark_text
    :param kwargs: additional keyword arguments for matplotlib.pyplot.text
    :return: None
    """

    if not watermark_text:
        if use_default:
            watermark_text: str = style.watermark_text
        else:
            return None

    if not style:
        style: Style = Style()

    # Accept either (x,y) placement OR a single scalar margin that will be used in both dimensions
    if isinstance(style.watermark_placement, (float, int)):
        position: Tuple[float, float] = (
            style.watermark_placement,
            style.watermark_placement,
        )
    elif len(style.watermark_placement) == 2:
        position: Tuple[float, float] = style.watermark_placement
    else:
        raise ValueError(f"Unable to interpret watermark {style.watermark_placement=}")

    ax = plt.gca()
    x_coordinate: float = margin_calc(position[0], plt.xlim(), ax.xaxis.get_scale())
    y_coordinate: float = margin_calc(position[1], plt.ylim(), ax.yaxis.get_scale())
    plt.text(
        x_coordinate,
        y_coordinate,
        watermark_text,
        fontsize=style.watermark_fontsize,
        c=style.watermark_color,
        **kwargs,
    )


def apply_plot_lines(
    axhline: Union[float, List[float]] = None,
    axvline: Union[float, List[float]] = None,
    style: Style = None,
    **kwargs,
) -> None:
    """
    Helper function to apply horizontal and vertical lines to a plot

    :param axhline: horizontal line(s) to plot
    :param axvline: vertical line(s) to plot
    :param style: config Style object
    :param kwargs: additional keyword arguments for matplotlib.pyplot.axhline and matplotlib.pyplot.axvline
    """
    if style is None:
        style: Style = Style()
    if axhline is not None:
        if isinstance(axhline, (int, float)):
            axhline = [axhline]
        for line in axhline:
            plt.axhline(
                line,
                color=style.axhline_color,
                linestyle=style.axhline_linestyle,
                linewidth=style.axhline_linewidth,
                **kwargs,
            )
    if axvline is not None:
        if isinstance(axvline, (int, float)):
            axvline = [axvline]
        for line in axvline:
            plt.axvline(
                line,
                color=style.axvline_color,
                linestyle=style.axvline_linestyle,
                linewidth=style.axvline_linewidth,
                **kwargs,
            )


##################
# Core visualize
# functionality
##################


def visualize_1d_distribution(
    data: Any,
    xlabel: str = "",
    ylabel: str = "counts",
    title: str = "",
    log_axes: str = "",
    style: Style = None,
    watermark: str = "",
    multi: bool = None,
    legend_strings: Union[Tuple[str], List[str]] = None,
    xlim: Any = None,
    ylim: Any = None,
    describe: bool = False,
    plot_mean: Union[bool, str] = False,
    plot_median: Union[bool, str] = False,
    show: bool = False,
    axhline: Union[float, List[float]] = None,
    axvline: Union[float, List[float]] = None,
    annotate_raw_counts: bool = False,
    downsample_viz_by_N_rows: int = 1,
    **kwargs,
) -> plt.Figure:
    """Core function for visualizing 1-dimensional distribution(s)

    :param data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param log_axes: which axes should be on a log scale, e.g. 'x' or ['x','y'] or 'xy'
    :param style: configuration object (optional)
    :param watermark: watermark text
    :param multi: flag to specify whether multiple traces are desired, if the automatic inference is incorrect
    :param legend_strings: override legend strings
    :param xlim: optional bound for the x-axis
    :param ylim: optional bound for the y-axis
    :param describe: adds stats from pandas.DataFrame.describe() as a watermark
    :param plot_mean: plots the mean as a vertical line
    :param plot_median: plots the median as a vertical line
    :param show: set to true to trigger plt.show()
    :param axhline: plots a horizontal line at the specified value(s)
    :param axvline: plots a vertical line at the specified value(s)
    :param downsample_viz_by_N_rows: downsample the data by this factor for visualization purposes
    :param kwargs: additional keyword arguments for matplotlib.pyplot.hist()
    :return: figure handle for the plot
    """
    # Set style. Overrides: kwargs > style input > Style() defaults
    original_kwargs: dict = kwargs.copy()
    config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})
    kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if k not in config.dict()}

    # If not specified, try to ascertain whether one or multiple data sets are being provided
    if multi or ((multi is None) and looks_like_list_of_lists(data)):
        config.color = None
        kwargs.setdefault("alpha", config.multi_hist_alpha)
    else:
        data: List[Any] = [data]

    # Plot the data and (and best fits if applicable)
    figure_handle: plt.Figure = plt.figure(facecolor=config.facecolor, figsize=config.figsize)
    plt.axes().set_prop_cycle(config.cycler)
    for data_set in data:
        # plotting on log x-axis requires special pre-treatment (log-distributed bin edges)
        if "x" in log_axes:
            if bins := kwargs.get("bins"):
                if not isinstance(bins, (int, float)):
                    hist_bins = bins
                else:
                    hist_bins = np.logspace(np.log10(min(data_set)), np.log10(max(data_set)), bins)
            else:
                hist_bins = np.logspace(
                    np.log10(min(data_set)),
                    np.log10(max(data_set)),
                    config.histogram_bins,
                )
        else:
            hist_bins = kwargs.get("bins")
        counts, bins, patches = plt.hist(
            data_set[::downsample_viz_by_N_rows],
            color=config.color,
            bins=hist_bins,
            **{k: v for k, v in kwargs.items() if not any(x_ in k for x_ in ["plot", "bins"])},
        )

        # Add text annotations for each bar
        if annotate_raw_counts:
            for count, patch in zip(counts, patches):
                plt.text(
                    patch.get_x() + patch.get_width() / 2,  # X position
                    patch.get_height(),  # Y position
                    f"{int(count)}",  # Text (count)
                    ha="center",  # Center alignment
                    va="bottom",  # Place text at the bottom of the count
                )

        if describe:
            description: pd.Series = pd.DataFrame({"_data_": data_set})["_data_"].describe()
            stat_str: str = dict_pretty(
                {stat_name: description[stat_name] for stat_name in description.keys().tolist()}
            )
            watermark += f"\n--- stats ---\n{stat_str}"
            if "watermark_fontsize" not in original_kwargs:
                config.watermark_fontsize = 12
            if "watermark_placement" not in original_kwargs:
                config.watermark_placement = (0.05, 0.6)

        if plot_mean:
            line_color: str = plot_mean if isinstance(plot_mean, str) else config.median_linecolor
            plt.axvline(x=np.nanmean(data_set), color=line_color, linestyle="-")
        if plot_median:
            line_color: str = plot_mean if isinstance(plot_mean, str) else config.median_linecolor
            plt.axvline(x=np.nanmedian(data_set), color=line_color, linestyle="--")

    # Make sure that the axes facecolor matches the figure facecolor
    plt.gca().set(facecolor=config.facecolor)

    # Adjust view & style where applicable
    xlabel_buffer: str = config.translate(xlabel, missing_response="return_input")
    ylabel_translated: str = config.translate(ylabel, missing_response="return_input")
    if kwargs.get("cumulative"):
        ylabel_buffer: str = f"cumulative {ylabel_translated}"
    else:
        ylabel_buffer: str = ylabel_translated
    if kwargs.get("density"):
        ylabel_buffer += " (density)"
    apply_plot_labels(xlabel=xlabel_buffer, ylabel=ylabel_buffer, title=title, style=config)
    if legend_strings:
        plt.legend(legend_strings, fontsize=config.legend_fontsize)
    apply_plot_lines(axhline=axhline, axvline=axvline, style=config)
    adjust_axes(log_axes=log_axes, style=config, xlim=xlim, ylim=ylim)
    apply_watermark(watermark, style=config)
    if show:
        plt.show()
    return figure_handle


def visualize_x_y(
    x_data: Any,
    y_data: Any,
    xlabel: str = "",
    ylabel: str = "",
    title: str = "",
    cumulative: str = "",
    log_axes: Union[str, List[str]] = "",
    types: Union[str, List[str]] = "scatter",
    style: Style = None,
    watermark: str = "",
    multi: bool = None,
    legend_strings: Union[Tuple[str], List[str]] = None,
    loc: str = None,
    xlim: Any = None,
    ylim: Any = None,
    plot_best_fit: Union[bool, int] = False,
    rolling_mean_width: int = None,
    rolling_median_width: int = None,
    show_colorbar: bool = False,
    log_norm_colors: bool = False,
    colorbar_label: str = None,
    x_axis_human_tick_labels: bool = False,
    x_axis_formatter: str = "%Y-%m-%d",
    show: bool = False,
    axvline: Union[List[float], float] = None,
    axhline: Union[List[float], float] = None,
    downsample_viz_by_N_rows: int = 1,
    infer_color_downsample: bool = True,
    **kwargs,
) -> plt.Figure:
    """Core function for visualizing 2-dimensional data sets

    :param x_data: can be an array-like object or a list of array-like objects (for multiple traces)
    :param y_data: can be an array-like object or a list of array-like objects (for multiple traces)
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param cumulative: which axes to make cumulative, e.g. 'x' or ['x','y'] or 'xy'
    :param log_axes: which axes should be on a log scale, e.g. 'x' or ['x','y'] or 'xy'
    :param types: which types of plot to make, currently supports: 'scatter' and 'plot'
    :param style: configuration object (optional)
    :param watermark: watermark text
    :param multi: flag to specify whether multiple traces are desired, if the automatic inference is incorrect
    :param legend_strings: override legend strings
    :param loc: location of the legend (matplotlib.pyplot convention)
    :param plot_best_fit: whether to plot the best fit lines (specify degree, or pass True for degree 1)
    :param xlim: optional bound for the x-axis
    :param ylim: optional bound for the y-axis
    :param rolling_median_width: window width for rolling average taken by df.y.rolling(rolling_mean_width).mean()
    :param rolling_mean_width: window width for rolling average taken by df.y.rolling(rolling_median_width).median()
    :param colorbar_label: optional label for the colorbar
    :param log_norm_colors: set to True to normalize the colorbar scale
    :param show_colorbar: set to True to show colorbar
    :param x_axis_human_tick_labels: set to True to convert numeric values along the x-axis to human-readable timestamps
    :param x_axis_formatter: format string for the x-axis if human-readable
    :param show: set to true to trigger plt.show()
    :param axvline: list of x-values to draw vertical lines at
    :param axhline: list of y-values to draw horizontal lines at
    :param downsample_viz_by_N_rows: downsample the data by this factor (for speed)
    :param infer_color_downsample: set to False to disable automatic downsampling of the color data
    :param kwargs: additional keyword arguments for matplotlib.pyplot.scatter()
    :return: figure handle for the plot
    """
    # Set style. Overrides: kwargs > style input > Style() defaults
    config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})
    kwargs: Dict[str, Any] = {
        k: v for k, v in kwargs.items() if (k not in config.dict()) or (k == "cmap")
    }  # allowlisting cmap is a hacky fix until the config handling is refactored

    # Set sampling rate to 1 (no downsampling) if None is passed in for downsample_viz_by_N_rows
    if not downsample_viz_by_N_rows:
        downsample_viz_by_N_rows: int = 1

    x_data: List[Any] = to_list_if_other_array(x_data)
    y_data: List[Any] = to_list_if_other_array(y_data)

    # If not specified, try to ascertain whether one or multiple data sets are being provided
    if multi or ((multi is None) and looks_like_list_of_lists(y_data)):
        config.color = None
    else:
        x_data: List[Any] = [x_data]
        y_data: List[Any] = [y_data]

    # Plot the data and (and best fits if applicable)
    figure_handle: plt.Figure = plt.figure(facecolor=config.facecolor, figsize=config.figsize)
    plt.axes().set_prop_cycle(config.cycler)
    scatter_handles: List[Any] = []

    includes_line_plot: bool = False
    for i, data_set in enumerate(zip(x_data, y_data)):
        if "x" in cumulative:
            x_array: np.ndarray = np.cumsum(data_set[0])
        else:
            x_array: np.ndarray = np.asarray(data_set[0])
        if "y" in cumulative:
            y_array: np.ndarray = np.cumsum(data_set[1])
        else:
            y_array: np.ndarray = np.asarray(data_set[1])

        # Convert to datetime for plotting if intending to use human-readable format
        if x_axis_human_tick_labels and any(not isinstance(x, datetime.datetime) for x in x_array):
            if any(np.isnan(x_array)):
                raise ValueError(
                    "Cannot convert NaN to datetime. Use numeric x-axis labels or filter NaNs upstream."
                )
            x_array: List[datetime.datetime] = [datetime.datetime.fromtimestamp(ts) for ts in x_array]

        if "scatter" in types:
            if log_norm_colors:
                kwargs.setdefault("norm", matplotlib.colors.LogNorm())
            if kwargs.get("c") is None:
                kwargs.setdefault("color", config.color)
            else:
                if (downsample_viz_by_N_rows > 1) and infer_color_downsample:
                    c_data: Any = kwargs["c"]
                    # Logic to infer whether or not we should try to downsample the color data
                    if (
                        not isinstance(c_data, (str, float, int))
                        and hasattr(c_data, "__len__")
                        and len(c_data) not in (0, 1, 3)
                        and len(c_data) == len(x_array)
                        and len(c_data) == len(y_array)
                    ):
                        kwargs["c"]: Any = c_data[::downsample_viz_by_N_rows]
            scatter_handles.append(
                plt.scatter(
                    x_array[::downsample_viz_by_N_rows],
                    y_array[::downsample_viz_by_N_rows],
                    config.markersize,
                    **kwargs,
                )
            )

        if any(x in types for x in ["plot", "line"]):
            includes_line_plot: bool = True
            p = plt.plot(
                x_array[::downsample_viz_by_N_rows],
                y_array[::downsample_viz_by_N_rows],
                color=config.color,
                linewidth=config.linewidth,
            )

        # Make sure that the axes facecolor matches the figure facecolor
        plt.gca().set(facecolor=config.facecolor)

        if plot_best_fit:
            if includes_line_plot:
                color: Any = p[0].get_color()  # noqa: misses earlier assignment
            else:
                color: Any = None
            if isinstance(plot_best_fit, int):
                degree: int = plot_best_fit
            else:
                degree: int = 1
            plot_best_fit_line(x_array, y_array, degree=degree, color=color, style=config)

        if rolling_mean_width or rolling_median_width:
            df: pd.DataFrame = pd.DataFrame({"x": x_array, "y": y_array})
            df.sort_values(by="x", ascending=True, inplace=True, ignore_index=True)
            if rolling_mean_width:
                plt.plot(
                    df.x,
                    df.y.rolling(rolling_mean_width).mean(),
                    color=config.mean_linecolor,
                    linewidth=config.mean_linewidth,
                    linestyle=config.mean_linestyle,
                )
            if rolling_median_width:
                plt.plot(
                    df.x,
                    df.y.rolling(rolling_median_width).median(),
                    color=config.median_linecolor,
                    linewidth=config.median_linewidth,
                    linestyle=config.median_linestyle,
                )

    # Adjust view & style where applicable
    xlabel_buffer: str = config.translate(xlabel, missing_response="return_input")
    if "x" in cumulative:
        xlabel_buffer += " (cumulative)"
    ylabel_buffer: str = config.translate(ylabel, missing_response="return_input")
    if "y" in cumulative:
        ylabel_buffer += " (cumulative)"
    apply_plot_labels(xlabel=xlabel_buffer, ylabel=ylabel_buffer, title=title, style=config)
    if show_colorbar or colorbar_label:
        cbar: plt.colorbar.Colorbar = plt.colorbar()
        if colorbar_label:
            cbar.set_label(colorbar_label, rotation=90, fontsize=config.label_fontsize)
    if legend_strings:
        if "scatter" in types:
            plt.legend(scatter_handles, legend_strings, fontsize=config.legend_fontsize, loc=loc)
        elif includes_line_plot:  # scatter legend overrides plot legend for now
            plt.legend(legend_strings, fontsize=config.legend_fontsize, loc=loc)
    apply_plot_lines(axhline=axhline, axvline=axvline, style=config)
    adjust_axes(
        log_axes=log_axes,
        style=config,
        xlim=xlim,
        ylim=ylim,
        x_axis_human_tick_labels=x_axis_human_tick_labels,
        x_axis_formatter=x_axis_formatter,
    )
    apply_watermark(watermark, style=config)
    if show:
        plt.show()
    return figure_handle


def visualize_1d_distribution_interpreter(*args, **kwargs) -> plt.Figure:
    """Wrapper for visualize_x_y that can take as inputs:
    + arrays
    + lists
    + data frame (+ feature names to plot)
    """
    config: Style = kwargs.get("style", Style())
    data: Any = None
    if (num_positional_arguments := len(args)) == 1:
        data = args[0]

    # Received two positional inputs (interpreted as x_data & y_data arrays, or a list of such arrays)
    elif num_positional_arguments == 2:
        if isinstance(args[0], pd.DataFrame):
            if isinstance(args[1], list):
                data = [args[0].loc[:, x].tolist() for x in args[1]]
                kwargs.setdefault("legend_strings", [config.translate(x) for x in args[1]])
            elif isinstance(args[1], str):
                data = args[0].loc[:, args[1]].tolist()
                kwargs.setdefault("xlabel", config.translate(args[1]))

    # Pass through to visualize_1d_distribution
    if data is not None:
        return visualize_1d_distribution(data=data, **kwargs)
    else:
        raise ValueError(
            "Issue encountered in visualize_1d_distribution_interpreter(), could not interpret the inputs"
        )


def visualize_x_y_input_interpreter(*args, **kwargs) -> plt.Figure:
    """Wrapper for visualize_x_y that can take as inputs:
    + arrays
    + lists
    + dictionary
    + data frame (+ feature names to plot)
    + VectorMultiSet
    + VectorSequence
    """
    config: Style = kwargs.get("style", Style())
    x_data: list = []
    y_data: list = []
    legend_strings: List[str] = []
    # Received a single positional input (so each item value must contain both x & y data)
    if (num_positional_arguments := len(args)) == 1:
        if isinstance((solo_input := args[0]), dict):
            # Here we have a dictionary with key=name, value=[x_data, y_data]
            x_data = [z[0] for z in solo_input.values()]
            y_data = [z[1] for z in solo_input.values()]
            legend_strings += solo_input.keys()
        elif isinstance(solo_input, list) and looks_like_list_of_lists(solo_input):
            # infer [[x1, y1], [x2, y2], ..., [xN, yN]]
            x_data = [z[0] for z in solo_input]
            y_data = [z[1] for z in solo_input]
        else:
            raise ValueError(f"Unknown input type: {type(solo_input)}")

    # Received two positional inputs (interpreted as x_data & y_data arrays, or a list of such arrays)
    elif num_positional_arguments == 2:
        if isinstance(args[0], pd.DataFrame) and isinstance(args[1], list):
            # infer: fxn(dataframe, [[x1name, y1name], [x2name, y2name], ...]
            x_data = [args[0].loc[:, z[0]] for z in args[1]]
            y_data = [args[0].loc[:, z[1]] for z in args[1]]
        elif len(args[0]) == len(args[1]):
            # infer: [x1, x2, ..., xN] and [y1, y2, ..., yN]   OR   [x_vec1, x_vec2, ...] and [y_vec1, y_vec2, ...]
            x_data = args[0]
            y_data = args[1]
        elif (not looks_like_list_of_lists(args[0])) or (
            looks_like_list_of_lists(args[0]) and (len(args[0]) == 1)
        ):
            if len(args[1]) > 1:
                # infer: fxn([x_all], [y1, y2, ..., yN], ...); also accepts: fxn(x_all, [y1, y2, ..., yN], ...)
                x_data = [args[0]] * len(args[1])
                y_data = args[1]

    # 3 inputs: for now this only handles a single x & y feature.  # TODO: extend to arbitrary number of pairs or y-axes
    elif num_positional_arguments == 3:
        # Check that the 2nd and 3rd inputs are strings (feature names)
        if all(isinstance(element, str) for element in args[1:3]):
            if isinstance(args[0], pd.DataFrame):
                df: pd.DataFrame = args[0]
            else:
                # If first argument NOT a dataframe, try to convert. This allows plots for isthmuslib vector collections
                try:
                    df: pd.DataFrame = pd.DataFrame(**args[0].dict())
                except AttributeError as ae:
                    raise AttributeError(
                        f"Tried to cast input to dict & dataframe, but was incompatible. Raised: {ae}"
                    )
                    # Check that the requested attributes are in the resultant data frame
            if not all(x in args[0].keys() for x in args[1:3]):
                raise ValueError(
                    f"Keys {args[1]} & {args[2]} not both in dataframe with: {df.keys().tolist()}"
                )
        else:
            raise ValueError(f"Expected positional arguments for feature names, got {args[1]} and {args[2]}")
        x_data = df[args[1]]
        y_data = df[args[2]]
        kwargs.setdefault("xlabel", config.translate(args[1]))
        kwargs.setdefault("ylabel", config.translate(args[2]))

    # Only use the auto-generated legend strings if user did not specify legend_strings
    kwargs.setdefault("legend_strings", legend_strings)

    # Pass through to visualize_x_y
    if (len(x_data) > 0) and (len(y_data) == len(x_data)):
        return visualize_x_y(x_data=x_data, y_data=y_data, **kwargs)
    else:
        raise ValueError(
            "Issue encountered in visualize_x_y_input_interpreter(), could not interpret the inputs"
        )


def visualize_hist2d(
    x_data: Any,
    y_data: Any,
    xlabel: str = "",
    ylabel: str = "",
    title: str = "",
    style: Style = None,
    watermark: str = "",
    plot_best_fit: Union[bool, int] = False,
    xlim: Any = None,
    ylim: Any = None,
    show_colorbar: bool = True,
    colorbar_label: str = "counts",
    zscale: str = "linear",
    show: bool = False,
    **kwargs,
) -> plt.Figure:
    """Visualize a 2-dimensional histogram (basically a heatmap of counts)

    :param x_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param y_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param style: configuration object (optional)
    :param watermark: watermark text
    :param plot_best_fit: whether to plot the best fit lines (specify degree, or pass True for degree 1)
    :param xlim: optional bound for the x-axis
    :param ylim: optional bound for the y-axis
    :param zscale: whether the colormap should be 'linear' or 'log'
    :param colorbar_label: text label to place next to the colorbar
    :param show_colorbar: whether to include the colorbar on teh plot
    :param show: set to true to trigger plt.show()
    :param kwargs: additional keyword arguments for matplotlib.pyplot.hist2d()
    :return: figure handle for the plot
    """
    # Set style. Overrides: kwargs > style input > Style() defaults
    config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})
    kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if (k not in config.dict()) or (k == "cmap")}
    kwargs.setdefault("cmap", config.sequential_cmap)

    x_data: List[Any] = to_list_if_other_array(x_data)
    y_data: List[Any] = to_list_if_other_array(y_data)

    # Make the plot
    figure_handle: plt.Figure = plt.figure(facecolor=config.facecolor, figsize=config.figsize)
    if zscale.lower() == "log":
        kwargs.setdefault("norm", matplotlib.colors.LogNorm())
    elif zscale.lower() != "linear":
        raise ValueError(f"zscale should be 'linear' or 'log' but received: {zscale}")
    plt.hist2d(x_data, y_data, **kwargs)
    if plot_best_fit:
        if isinstance(plot_best_fit, int):
            degree: int = plot_best_fit
        else:
            degree: int = 1
        plot_best_fit_line(x_data, y_data, degree=degree, color="k", style=config)

    # Adjust view & style where applicable
    plt.xlim(xlim)
    plt.ylim(ylim)
    apply_plot_labels(xlabel=xlabel, ylabel=ylabel, title=title, style=config)
    apply_watermark(watermark, style=config)
    if show_colorbar:
        cbar: plt.colorbar.Colorbar = plt.colorbar()
        cbar.set_label(colorbar_label, rotation=90, fontsize=config.label_fontsize)
    if show:
        plt.show()
    return figure_handle


def visualize_surface(
    x_data,
    y_data,
    z_data,
    xlabel: str = "",
    ylabel: str = "",
    title: str = "",
    xlim: Any = None,
    ylim: Any = None,
    style: Style = None,
    watermark: str = None,
    y_axis_ascending: bool = True,
    show: bool = False,
    **kwargs,
) -> plt.Figure:
    """Plots a surface (note: seaborn heatmap has somewhat rigid requirements for data completeness)

    :param x_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param y_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param z_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param xlim: optional bound for the x-axis
    :param ylim: optional bound for the y-axis
    :param style: configuration object (optional)
    :param watermark: watermark text
    :param y_axis_ascending: sort the axis low to high values bottom to top
    :param show: set to true to trigger plt.show()
    :param kwargs: additional keyword arguments for seaborn heatmap()
    :return: figure handle for the plot
    """
    # Set style. Overrides: kwargs > style input > Style() defaults
    config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})

    kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if (k not in config.dict()) or (k == "cmap")}
    kwargs.setdefault("cmap", config.sequential_cmap)

    x_data: List[Any] = to_list_if_other_array(x_data)
    y_data: List[Any] = to_list_if_other_array(y_data)
    z_data: List[Any] = to_list_if_other_array(z_data)

    # Initial data wrangling
    df: pd.DataFrame = pd.DataFrame({"x": x_data, "y": y_data, "z": z_data})
    pivoted: pd.DataFrame = df.pivot(index="y", columns="x", values="z")

    # Make the plot
    figure_handle: plt.Figure = plt.figure(facecolor=config.facecolor, figsize=config.figsize)
    ax: plt.Axes = sns.heatmap(pivoted, **kwargs)

    # Adjust view & style where applicable
    plt.xlim(xlim)
    plt.ylim(ylim)

    apply_plot_labels(xlabel=xlabel, ylabel=ylabel, title=title, style=config)
    apply_watermark(watermark, style=config)
    if y_axis_ascending:
        ax.invert_yaxis()
    if show:
        plt.show()
    return figure_handle


def visualize_embedded_surface(
    x_data,
    y_data,
    z_data,
    xlabel: str = "",
    ylabel: str = "",
    title: str = "",
    xlim: Any = None,
    ylim: Any = None,
    style: Style = None,
    show_colorbar: bool = True,
    log_norm_colors: bool = True,
    show: bool = False,
    **kwargs,
) -> plt.Figure:
    """Plots a 2D surface in 3D

    :param x_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param y_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param z_data: can be an list-like object or a list of list-like objects (for multiple traces)
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param xlim: optional bound for the x-axis
    :param ylim: optional bound for the y-axis
    :param style: configuration object (optional)
    :param kwargs: additional keyword arguments for seaborn heatmap()
    :param log_norm_colors: set to True to normalize the colorbar scale
    :param show_colorbar: set to True to show colorbar
    :param show: set to true to trigger plt.show()
    :return: figure handle for the plot
    """
    # Set style. Overrides: kwargs > style input > Style() defaults
    config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})
    kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if (k not in config.dict()) or (k == "cmap")}
    kwargs.setdefault("cmap", config.sequential_cmap)

    x_data: List[Any] = to_list_if_other_array(x_data)
    y_data: List[Any] = to_list_if_other_array(y_data)
    z_data: List[Any] = to_list_if_other_array(z_data)

    # Make the plot
    figure_handle: plt.Figure = plt.figure(facecolor=config.facecolor, figsize=config.figsize)
    ax = Axes3D(figure_handle)
    if log_norm_colors:
        kwargs.setdefault("norm", matplotlib.colors.LogNorm())
    figure_surface = ax.plot_trisurf(x_data, y_data, z_data, **kwargs)

    # Adjust view & style where applicable
    plt.xlim(xlim)
    plt.ylim(ylim)

    if show_colorbar:
        figure_handle.colorbar(figure_surface, shrink=0.5, aspect=5)
    apply_plot_labels(xlabel=xlabel, ylabel=ylabel, title=title, style=config)
    if show:
        plt.show()
    return figure_handle


def surface_from_dataframe(
    df: pd.DataFrame,
    x_col_name: str = "x",
    y_col_name: str = "y",
    z_col_name: str = "z",
    **kwargs,
) -> plt.Figure:
    """Extremely thin wrapper around plot_surface() that extracts the data from a dataframe"""
    col_names: List[str] = [x_col_name, y_col_name, z_col_name]
    if any(element not in (keys := df.keys().tolist()) for element in col_names):
        raise ValueError(f"Could not find all of {col_names} in data frame keys: {keys}")
    return visualize_surface(df[x_col_name], df[y_col_name], df[z_col_name], **kwargs)


def dict_to_bar_chart(
    data,
    xlabel: str = "",
    ylabel="",
    title: str = "",
    cmap="tab10",
    yscale: str = "linear",
    grid: str = "xy",
    show: bool = True,
    figsize: Tuple[float, float] = (10.0, 6.0),
    facecolor: str = "white",
    xtick_rotation: int = 0,
    xtick_fontsize: int = 12,
    title_fontsize: int = 16,
    xlabel_fontsize: int = 14,
    ylabel_fontsize: int = 14,
) -> plt.Figure:
    """
    Plots a bar chart given a dict with string keys and numeric values.

    :param data: dictionary of data to plot
    :param xlabel: label text for the x-axis
    :param ylabel: label text for the y-axis
    :param title: title text
    :param cmap: color map for bars (cycles)
    :param yscale: scale for y-axis
    :param grid: whether to show grid lines (can be 'x' or 'y', or 'xy'/'on' for both)
    :param show: whether to show the plot
    :param figsize: size of the figure
    :param facecolor: facecolor for the figure
    :param xtick_rotation: rotation for x-axis tick labels
    :param xtick_fontsize: fontsize for x-axis tick labels
    :param title_fontsize: fontsize for the title
    :param xlabel_fontsize: fontsize for the x-axis label
    :param ylabel_fontsize: fontsize for the y-axis label
    :return: figure handle for the plot
    """

    # Prepare color cycle
    colormap = plt.get_cmap(cmap)
    colors = cycle(colormap.colors)  # Cycling the colors

    # Create the figure
    f: plt.Figure = plt.figure(figsize=figsize, facecolor=facecolor)
    for key, value in data.items():
        plt.bar(key, value, color=next(colors))

    # Adjust view & style where applicable
    plt.yscale(yscale)
    plt.xlabel(xlabel, fontsize=xlabel_fontsize)
    plt.ylabel(ylabel, fontsize=ylabel_fontsize)
    plt.title(title, fontsize=title_fontsize)
    plt.xticks(rotation=xtick_rotation, fontsize=xtick_fontsize)

    # Modify grid if desired
    if "x" in grid.lower() or "on" in grid.lower():
        plt.grid(axis="x", linestyle="--", alpha=0.7)
    if "y" in grid.lower() or "on" in grid.lower():
        plt.grid(axis="y", linestyle="--", alpha=0.7)

    # Show the plot
    if show:
        plt.show()

    return f


###################
# Aliases to mirror
# matplotlib.pyplot
###################

hist: Callable[[Any], plt.Figure] = visualize_1d_distribution_interpreter
hist2d: Callable[[Any], plt.Figure] = visualize_hist2d
surface: Callable[[Any], plt.Figure] = visualize_surface


def scatter(*args, **kwargs) -> plt.Figure:
    """Super thin wrapper for visualize(), to mimic matplotlib.plt interface"""
    return visualize_x_y_input_interpreter(*args, **kwargs, types="scatter")


def plot(*args, **kwargs) -> plt.Figure:
    """Super thin wrapper for visualize(), to mimic matplotlib.plt interface"""
    return visualize_x_y_input_interpreter(*args, **kwargs, types="plot")



import time as time
from typing import List, Any, Tuple, Callable, Dict, Union, Iterable, Optional
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# import stumpy
from matplotlib.patches import Rectangle
from .config import Style
from .utils import (
    PickleUtils,
    Rosetta,
    make_dict,
    get_num_workers,
    process_queue,
    dict_pretty,
    zero_mean_unit_variance,
    convert_dtypes_subset,
    df_to_any,
    df_read_any,
    convert_dtypes_automatically_subset,
    convert_dtypes_advanced,
    machine_time,
)
from .data_quality import basis_quality_checks, basis_quality_plots, fill_ratio
from copy import deepcopy
import statsmodels.api as sm
from .plotting import (
    visualize_x_y,
    visualize_1d_distribution,
    visualize_surface,
    visualize_embedded_surface,
    visualize_hist2d,
    visualize_x_y_input_interpreter,
)
import pathlib
from pydantic import BaseModel
from tqdm.auto import tqdm

# import matrixprofile
import math
import sys


class SVD(BaseModel):
    u: Any
    s: Any
    vh: Any


class VectorMultiset(PickleUtils, Style, Rosetta):
    """A set of vectors (which may or may not be ordered)"""

    data: Optional[Any] = None  # needs to be dataframe or path/str to CSV file
    name_root: Optional[str] = None
    svd: Optional[SVD] = None

    class Config:
        arbitrary_types_allowed = True

    def __len__(self) -> int:
        """The length of the vector set is the length of the data frame"""
        return len(self.data)

    def __add__(self, other):
        """Adding items derived from MultiSets concatenates their dataframes (keeping metadat from left side)"""
        to_return = deepcopy(self)
        to_return.data = pd.concat([self.data, other.data])
        return to_return

    def head(self, *args, **kwargs) -> pd.DataFrame:
        """For convenience, alias to pandas method DataFrame.head()"""
        return self.data.head(*args, **kwargs)

    def tail(self, *args, **kwargs) -> pd.DataFrame:
        """For convenience, alias to pandas method DataFrame.tail()"""
        return self.data.tail(*args, **kwargs)

    def values(self, feature: str, cumulative: bool = False, *args) -> List[Any]:
        """Retrieves a particular data feature by attribute name. Additional args unpack deeper

        :param feature: name of feature to retrieve
        :param cumulative: whether to apply a cumulative sum (default = False)
        :return: Extracted data
        """
        if feature not in self.data.keys():
            raise ValueError(f"Feature {feature} not in known keys: {self.data.keys().tolist()}")
        values: pd.Series = self.data.loc[:, feature]
        if args:
            for unpack_next in args:
                values: List[Any] = [x.__getattribute__(unpack_next) for x in values]
        if cumulative:
            return np.cumsum(values).tolist()
        return values.tolist()

    def dict_pretty(self, max_length: int = None, previews: bool = False) -> str:
        """
        Helper function to make a preview string (that doesn't become excessively long with data present)

        :param max_length: max length of value
        :param previews: whether to include previews for longer values
        :return: dictionary summary string
        """
        return dict_pretty(
            self.dict(),
            max_length=max_length if max_length else self.dict_pretty_max_length,
            previews=previews if previews is not None else self.dict_pretty_previews,
        )

    def keys(self) -> List[str]:
        """Returns the dataframe columns as a list of strings"""
        return self.data.keys().tolist()

    def convert_dtypes_subset(self, new_col_types: Dict[str, Union[type, str]] = None) -> None:
        """Converts the data types of a subset of columns, wraps utils.convert_dtypes_subset

        :param cols: columns to convert (if None, converts all)
        """
        self.data = convert_dtypes_subset(self.data, new_col_types)

    def convert_dtypes_automatically_subset(self, cols: List[str] = None) -> None:
        """Converts the data types of a subset of columns, wraps utils.convert_dtypes_subset

        :param cols: columns to convert (if None, converts all)
        """
        self.data = convert_dtypes_automatically_subset(self.data, cols)

    def convert_dtypes_advanced(self, **kwargs) -> None:
        """Converts the data types of a subset of columns, wraps utils.convert_dtypes_advanced"""
        self.data = convert_dtypes_advanced(self.data, **kwargs)

    ################
    # I/O
    ################

    def to_df(self) -> pd.DataFrame:
        """Trivial helper function

        :return: hands back the self.data data frame
        """
        return self.data

    def from_dataframe(self, data_frame: pd.DataFrame, inplace: bool = True, **kwargs) -> Union[Any, None]:
        """Makes an VectorMultiset from a pandas data frame

        :param data_frame: data frame to import
        :param inplace: import inplace (default) or return the result
        :return: data frame if not inplace
        """
        if inplace:
            self.data = data_frame
            for key, value in kwargs.items():
                self.__setattr__(key, value)
        else:
            return self.__class__(data=data_frame, **kwargs)

    def to_csv(self, file_path: Union[str, pathlib.Path] = None) -> None:
        """Saves the data as a CSV file (note: this drops the name_root)

        :param file_path: path to write the file
        """
        if not file_path:
            file_path: str = f"default_filename_{time.time():.0f}.csv"
        self.data.to_csv(file_path)

    def read_csv(
        self, file_path: Union[str, pathlib.Path], inplace: bool = True, **kwargs
    ) -> Union[Any, None]:
        """Reads data from a CSV file

        :param file_path: file to read
        :param inplace: import inplace (default) or return the result
        """
        data: pd.DataFrame = pd.read_csv(file_path)
        if inplace:
            self.data = data
            for key, value in kwargs.items():
                self.__setattr__(key, value)
        else:
            return self.__class__(data=data, **kwargs)

    def to_feather(self, file_path: Union[str, pathlib.Path], **kwargs) -> None:
        """Saves the data as a feather file (note: this drops the name_root). keyword

        :param file_path: path to write the file
        """
        self.data.to_feather(file_path, **kwargs)

    def read_feather(
        self, file_path: Union[str, pathlib.Path], inplace: bool = True, **kwargs
    ) -> Union[Any, None]:
        """Reads data from a CSV file

        :param file_path: file to read
        :param inplace: import inplace (default) or return the result
        """
        data: pd.DataFrame = pd.read_feather(file_path)
        if inplace:
            self.data = data
            for key, value in kwargs.items():
                self.__setattr__(key, value)
        else:
            return self.__class__(data=data, **kwargs)

    def to_any(
        self, paths: Union[List[Union[str, pathlib.Path]], Union[str, pathlib.Path]], **kwargs
    ) -> None:
        """Saves the data as csv, json, feather, parquet, pickle, or hdf (inferred from file extension)

        :param paths: path(s) to write the file
        """
        df_to_any(self.data, paths, **kwargs)

    def read_any(self, file_path: Union[str, pathlib.Path], inplace: bool = True, **kwargs) -> Any:
        df: pd.DataFrame = df_read_any(file_path, **kwargs)
        if inplace:
            self.data = df
        else:
            return self.__class__(data=df, **kwargs)

    ################
    # Visualizations
    ################

    def visualize_x_y(
        self,
        x: Union[str, List[Any], Any],
        y: Union[str, List[Any], Any],
        c: str = None,
        cumulative: Union[str, List[str]] = "",
        **kwargs,
    ) -> plt.Figure:
        """Visualize in two dimensions

        :param x: name of the x-axis data feature (or the data itself)
        :param y: name of the y-axis data feature (or the data itself)
        :param cumulative: Which (if any) dimensions should be cumulative, e.g. 'x' or ['x','y'] or 'xy'
        :param kwargs: additional kwargs for isthmuslib.visualize_x_y, passed through to matplotlib.pyplot.scatter()
        :return: figure handle for the plot
        """
        if isinstance(x, str):
            x_data: List[Any] = self.values(x, cumulative="x" in cumulative)
            kwargs.setdefault("xlabel", self.translate(x))
        else:
            x_data: Any = x

        if isinstance(y, str):
            y_data: List[Any] = self.values(y, cumulative="y" in cumulative)
            kwargs.setdefault("ylabel", self.translate(y))
        else:
            y_data: Any = y

        # Misc style
        kwargs.setdefault("title", self.translate(self.name_root))
        if (c is not None) and (isinstance(c, str)) and (c in self.data.keys()):
            kwargs["c"] = self.data.loc[:, c].tolist()
        else:
            kwargs["c"] = c
        return visualize_x_y(x_data, y_data, **kwargs)

    def viz2d(self, *args, **kwargs) -> plt.Figure:
        """
        For convenience, allow direct access to visualize_x_y_input_interpreter

        :param args: visualize_x_y_input_interpreter positional inputs
        :param kwargs: visualize_x_y_input_interpreter keyword argument inputs
        :return: figure handle
        """
        kwargs.setdefault("style", Style(**self.dict()))
        return visualize_x_y_input_interpreter(*args, **kwargs)

    def scatter(self, *args, **kwargs) -> plt.Figure:
        """Creates a 2D scatter plot of x and y data (wraps visualize_x_y)"""
        if isinstance(self, VectorSequence) and (len(args) == 1):
            args: tuple = (
                self.basis_col_name,
                args[0],
            )  # sequences we can infer the basis for the x-axis
        return self.visualize_x_y(types="scatter", *args, **kwargs)

    def plot(self, *args, **kwargs) -> plt.Figure:
        """Creates a 2D line plot of x and y data (wraps visualize_x_y)"""
        if isinstance(self, VectorSequence) and (len(args) == 1):
            args: tuple = (
                self.basis_col_name,
                args[0],
            )  # sequences we can infer the basis for the x-axis
        if isinstance(self, SlidingWindowResults) and (len(args) == 1):
            raise ValueError(f"Plot requires x & y. Hint: You might want SlidingWindowResults.plot_results()")
        return self.visualize_x_y(types="plot", *args, **kwargs)

    def hist(self, col_name: str, **kwargs) -> plt.Figure:
        """Plot a histogram of data. Useful kwargs: [cumulative, density, bins]

        :param col_name: name of field to plot
        :return: figure handle
        """
        data: List[Any] = self.values(col_name)
        kwargs.setdefault("xlabel", self.translate(col_name))
        kwargs.setdefault("title", self.translate(self.name_root))
        return visualize_1d_distribution(data, style=Style(**self.dict()), **kwargs)

    def hist2d(self, *args, **kwargs) -> plt.Figure:
        """Creates a 2D histogram plot of x and y data (wraps visualize_hist2d)"""
        if isinstance(self, VectorSequence) and (len(args) == 1):
            args: tuple = (
                self.basis_col_name,
                args[0],
            )  # sequences we can infer the basis for the x-axis

        # If the first argument is a string, replace it with column data
        if isinstance(args[0], str):
            kwargs.setdefault("xlabel", self.translate(args[0]))
            args = self.data.loc[:, args[0]].tolist(), *args[1:]

        # If the second argument is a string, replace it with column data
        if isinstance(args[1], str):
            kwargs.setdefault("ylabel", self.translate(args[1]))
            extra_args = args[2:] if len(args) > 0 else tuple()
            args = args[0], self.data.loc[:, args[1]].tolist(), *extra_args

        kwargs.setdefault("title", self.name_root)
        return visualize_hist2d(*args, **kwargs)

    def surface(self, x_name: str, y_name: str, z_name: str, **kwargs) -> plt.Figure:
        """Plot a surface from the data. Useful kwargs: [cumulative, density, bins]

        :param x_name: name of the x-axis data feature
        :param y_name: name of the y-axis data feature
        :param z_name: name of the z-axis data feature (color)
        :param kwargs: additional keyword arguments for visualize_surface which passes to seaborn heatmap()
        :return: figure handle for the plot
        """
        if isinstance(self, VectorSequence) and (not x_name):
            x_name: str = self.basis_col_name
        kwargs.setdefault("xlabel", self.translate(x_name, missing_response="return_input"))
        kwargs.setdefault("ylabel", self.translate(y_name, missing_response="return_input"))
        kwargs.setdefault("title", self.translate(z_name, missing_response="return_input"))
        return visualize_surface(
            self.data.loc[:, x_name],
            self.data.loc[:, y_name],
            self.data.loc[:, z_name],
            **kwargs,
        )

    def visualize_embedded_surface(
        self, x_col_name: str, y_col_name: str, z_col_name: str, **kwargs
    ) -> plt.Figure:
        """Plots a surface (2D embedded in 3D) based on vector features"""
        return visualize_embedded_surface(
            self.data.loc[:, x_col_name],
            self.data.loc[:, y_col_name],
            self.data.loc[:, z_col_name],
            **kwargs,
        )

    ################
    # Statistics
    ################

    def correlation_matrix(self, **kwargs) -> pd.DataFrame:
        """
        Very thin wrapper around correlation matrix (using pandas corr() method)

        :param kwargs: additional keyword arguments for correlation_matrix (fed through to df.style.background_gradient)
        :return: styled pandas dataframe
        """
        return correlation_matrix(self.data, **kwargs)

    def singular_value_decomposition(
        self, cols: Union[str, List[str]] = None, cache_results: bool = False, **kwargs
    ) -> SVD:
        """Applies singular value decomposition

        :param cache_results: if True, saves u, s, and vh as the attribute 'svd' (an SVD class)
        :param cols: which data features to use
        :param kwargs: keyword arguments for SVD
        :return: u, s, vh arrays
        """
        if not cols:
            cols = self.data.keys().tolist()
        svd: SVD = singular_value_decomposition(self.data, cols, **kwargs)
        if cache_results:
            self.svd = svd
        return svd

    def to_surface_df(
        self, group_by_col_names: List[str], aggregation_method: str = "max", **kwargs
    ) -> pd.DataFrame:
        """Thin wrapper that aggregates the data to a surface, using pandas methods like mean, median, max, etc

        :param group_by_col_names: axes for the surface (i.e. how to bin the data)
        :param aggregation_method: how to combine multiple data points in the same bin
        :param kwargs: additional keyword arguments for pandas groupby method
        :return: data frame with combined according `aggregation_method` into bins of `group_by_col_names`
        """
        return getattr(self.data.groupby(by=group_by_col_names, **kwargs), aggregation_method)().reset_index(
            drop=False
        )

    def plot_projection_to_surface(
        self,
        surface_axes_names: Union[List[str], Tuple[str, str]],
        z_axis_name: str,
        aggregation_method: str = "max",
        **kwargs,
    ) -> plt.Figure:
        if len(surface_axes_names) != 2:
            raise ValueError(f"plot_projection is intended to be used with 2D surfaces")
        df_surface: pd.DataFrame = self.to_surface_df(
            group_by_col_names=surface_axes_names,
            aggregation_method=aggregation_method,
            **kwargs,
        )
        return visualize_x_y(
            df_surface.loc[:, surface_axes_names[0]],
            df_surface.loc[:, surface_axes_names[1]],
            c=df_surface.loc[:, z_axis_name].tolist(),
            xlabel=self.translate(surface_axes_names[0]),
            ylabel=self.translate(surface_axes_names[1]),
            colorbar_label=self.translate(z_axis_name),
        )

    # def feature_selection_univariate(
    #     self,
    #     target_feature_name: str,
    #     input_feature_names: List[str] = None,
    #     k_best: int = 3,
    #     normalize: bool = True,
    #     **kwargs,
    # ) -> any:
    #     """Feature selection of k best using univariate methods.
    #
    #     :param target_feature_name: column name of the target feature
    #     :param input_feature_names: column names of the input features
    #     :param k_best: how many features to return
    #     :param normalize: if True, scales input features to unit mean
    #     :param kwargs: additional keyword arguments passed through to scikit-learn SelectKBest
    #     :return: trimmed input feature data set
    #     """
    #     if not input_feature_names:
    #         input_feature_names = [x for x in self.data.keys() if x != target_feature_name]
    #     input_feature_data = self.data.loc[:, input_feature_names]
    #     if normalize:
    #         for fieldname in input_feature_data.keys():
    #             std_dev: float = float(np.std(this_field_data := self.data.loc[:, fieldname]))
    #             input_feature_data[fieldname] = [x / std_dev for x in this_field_data]
    #
    #     target_feature_data = self.data.loc[:, target_feature_name].to_numpy()
    #     return SelectKBest(chi2, k=k_best).fit_transform(input_feature_data, target_feature_data, **kwargs)

    def cast_to_numeric(
        self,
        columns: Union[List[str], str] = None,
        errors: str = "ignore",
        inplace: bool = True,
    ) -> Union[None, Any]:
        """
        Helper function that converts the data to numeric types

        :param columns: a string or list of strings of column names. If not provided, will try on all columns
        :param errors: passed to `pandas.to_numeric()`, options are: {‘ignore’, ‘raise’, ‘coerce’}
        :param inplace: whether to transform the data inplace (default) or return a fresh object
        """
        if not columns:
            columns = list(self.data.keys())
        if isinstance(columns, str):
            columns = [columns]

        df: pd.DataFrame = deepcopy(self.data)
        for col in columns:
            df[col] = pd.to_numeric(df.loc[:, col], errors=errors)

        if inplace:
            self.data = df
        else:
            return self.__class__(data=df, **{k: v for k, v in self.dict().items() if k != "data"})

    def __init__(self, **kwargs):

        # Intercept if 'data' is a path instead of a dataframe
        if isinstance(kwargs.get("data", None), (str, pathlib.Path)):
            data_input: Union[str, pathlib.Path] = kwargs.get("data")
            if ".csv" in str(data_input).lower():
                if pathlib.Path(data_input).exists():
                    kwargs["data"] = pd.read_csv(data_input)
                else:
                    raise ValueError(f"Input data string interpreted as a path does not exist:\n{data_input}")
            else:
                raise ValueError(
                    f"Input data appears to be a {type(data_input)} but is not a file with .csv extension"
                )

        super().__init__(**kwargs)

        if not kwargs.get("disable_auto_conversion_to_numeric", False) and (self.data is not None):
            self.cast_to_numeric()

    def drop_col_types(self, drop_types: Union[type, List[type]], inplace: bool = True) -> Union[Any, None]:
        """
        Helper function to drop columns of a particular type(s)

        :param drop_types: the types to be dropped
        :param inplace: whether the columns should be dropped inplace or if a new instance should be returned
        :return: None if inplace=True, otherwise a copy of the object with dropped columns
        """
        if not isinstance(drop_types, (list, tuple)):
            drop_types = [drop_types]
        target_cols: List[str] = [
            k for k in self.data.keys() if isinstance(self.data.loc[:, k][0], drop_types)
        ]
        if inplace:
            self.data.drop(columns=target_cols, inplace=True)
            return None
        else:
            to_return: Any = deepcopy(self)
            to_return.drop_col_types(drop_types=drop_types, inplace=True)
            return to_return

    def matrix_profile_univariate(
        self,
        col_names: Union[str, Iterable[str]],
        auto_locf: bool = False,
        suppress_nan_warning: bool = False,
        **kwargs,
    ) -> List[Tuple[Any, List[plt.Figure]]]:
        raise NotImplementedError(
            f"This function was deprecated when active support for matrixprofile was dropped"
        )

    #     """
    #     Very thin wrapper around matrixprofile with some warnings and opt-in preprocessing automation
    #
    #     :param col_names: a column name (or list of column names) to be profiled
    #     :param auto_locf:
    #     :param kwargs: keyword arguments passed through to matrixprofile.analyze()
    #     :return: List of outputs from matrixprofile.analyze()
    #     """
    #     queue: List[str] = [col_names] if isinstance(col_names, str) else col_names
    #     results: List[Tuple[Any, List[plt.Figure]]] = []
    #     try:
    #         for q in queue:
    #             ts: List[float] = self.data.loc[:, q].astype(float).tolist()
    #
    #             if not kwargs.get("preprocessing_kwargs") and any(
    #                 np.isnan(x) for x in ts
    #             ):
    #                 if hasattr(self, "auto_locf_params"):
    #                     auto_locf_params: Dict[str, Any] = self.auto_locf_params
    #                 else:
    #                     auto_locf_params: Dict[str, Any] = {
    #                         "window": 3,
    #                         "impute_method": "median",
    #                         "impute_direction": "forward",
    #                         "add_noise": False,
    #                     }
    #                 if auto_locf:
    #                     kwargs.setdefault("preprocessing_kwargs", auto_locf_params)
    #                 else:
    #                     if not suppress_nan_warning:
    #                         print(
    #                             """
    # Warning: matrix profiling data with NaNs and no preprocessing specified
    #          (potential anomalies or errors ahead)\n
    # > To suppress this warning in the future, specify `suppress_nan_warning=True`.
    # > One way to address NaNs is specifying a 'preprocessing_kwargs' dictionary, like:
    #     {'window': 3, 'impute_method': 'median', 'impute_direction': 'forward', 'add_noise': False}
    # > Alternatively, you can specify `auto_locf=True` to use the config presets:\n"""
    #                             + "".join(
    #                                 [
    #                                     f"       ~ {k}={v}\n"
    #                                     for k, v in auto_locf_params.items()
    #                                 ]
    #                             )
    #                         )
    #             results.append(matrixprofile.analyze(ts, **kwargs))
    #     except TypeError as e:
    #         s: str = f"""
    #         Encountered a TypeError in matrixprofile.analyze(). Hint: this might happen if you are
    #         trying to use a custom parameter like 'use_right_edge' with the standard matrixprofile
    #         package. If that is the case, try running:\n
    #           pip uninstall matrixprofile && pip install git+https://github.com/mitchellpkt/matrixprofile
    #         \nOriginal error: {e}"""
    #         print(s)
    #         raise TypeError(s)
    #
    #     return results

    def calculate_stumpy_profile_univariate(self, col_name: str, window_size: int, **kwargs) -> np.ndarray:
        """
        Very thin wrapper around stumpy's matrix profile method. NB: for use in MultiSet, user is responsible for order

        :param col_name: column name to profile
        :param window_size: sliding window size
        :param kwargs: keyword arguments for stumpy.stump()
        :return: numpy array with the profile
        """

        if "stumpy" not in sys.modules:
            try:
                import stumpy
            except ImportError as e:
                raise ImportError(
                    f"Could not import stumpy. Please install if compatible with your environment.\n{e}"
                )

        return stumpy.stump(self.data.loc[:, col_name].tolist(), window_size, **kwargs)

    def stumpy_profile_univariate(
        self, col_name: str, window_size: int, annotate_motif: bool = True, **kwargs
    ) -> plt.Figure:
        """
        Plot of data & matrix profile (with optional annotation). NB: for use in MultiSet, user is responsible for order

        :param col_name: name of the column to plot
        :param window_size: sliding window size
        :param annotate_motif: whether to highlight the main motif and its nearest neighbor
        :param kwargs: keyword arguments for plots and stumpy.stump()
        :return: figure handle of the plot
        """
        if "stumpy" not in sys.modules:
            try:
                import stumpy
            except ImportError as e:
                raise ImportError(
                    f"Could not import stumpy. Please install if compatible with your environment.\n{e}"
                )

        # Make the plot and profile
        figsize: Any = kwargs.pop("figsize", self.figsize)  # typically a 2-element tuple or list
        title: str = kwargs.pop("title", self.translate(self.name_root))
        profile: np.ndarray = self.calculate_stumpy_profile_univariate(col_name, window_size, **kwargs)
        minimum: float = float(np.nanmin(self.data.loc[:, col_name].tolist()))
        maximum: float = float(np.nanmax(self.data.loc[:, col_name].tolist()))
        fig, axs = plt.subplots(
            2,
            sharex="all",
            gridspec_kw={"hspace": 0},
            figsize=figsize,
            facecolor=self.facecolor,
        )
        plt.suptitle(title, fontsize=self.title_fontsize)
        axs[0].plot(self.data.loc[:, col_name].tolist(), color=kwargs.get("color", self.color))
        axs[0].set_ylabel(self.translate(col_name), fontsize=self.label_fontsize)
        axs[1].set_xlabel(self.translate("Index"), fontsize=self.label_fontsize)
        axs[1].set_ylabel("Matrix Profile", fontsize=self.label_fontsize)
        axs[1].plot(profile[:, 0], color=kwargs.get("color", self.color))

        # Annotate the main motif (and its nearest neighbor) if desired
        if annotate_motif:
            motif_idx = np.argsort(profile[:, 0])[0]
            nearest_neighbor_idx = profile[motif_idx, 1]
            rect = Rectangle((motif_idx, minimum), window_size, maximum, facecolor="palegoldenrod")
            axs[0].add_patch(rect)
            rect = Rectangle(
                (nearest_neighbor_idx, minimum),
                window_size,
                maximum,
                facecolor="palegoldenrod",
            )
            axs[0].add_patch(rect)
            axs[1].axvline(x=motif_idx, linestyle="dashed", color="k")
            axs[1].axvline(x=nearest_neighbor_idx, linestyle="dashed", color="k")

        return fig

    # Group the multiset by a given row
    def group_by_col(self, by: str) -> Dict[Any, pd.DataFrame]:
        """Helper function that returns a dictionary of individual dataframes separated by 'by'"""
        return {x: self.data.loc[self.data.loc[:, by] == x, :] for x in set(self.data.loc[:, by])}


class SlidingWindowResults(VectorMultiset):
    """Results from Sequence.sliding_window(), which is a VectorMultiset with extra context baked in"""

    window_width_col_name: str = "window_width"
    window_start_col_name: str = "window_start"

    def group_by_window_width(self) -> Dict[float, pd.DataFrame]:
        """Helper function that returns individual dataframes for each window width"""
        return self.group_by_col(by=self.window_width_col_name)

    def plot_results(self, col_name: str, legend_override: List[str] = None, **kwargs) -> plt.Figure:
        """Plot any sequence (based on the data frame column name)

        :param col_name: dataframe column to plot
        :param legend_override: specify custom legend keys if desired
        :return: figure handle
        """
        x_arrays: List[Any] = []
        y_arrays: List[Any] = []
        labels: List[Any] = []

        # For each `window_width` add we'll add a trace to the plot
        group_by_window_width: Dict[float, pd.DataFrame] = self.group_by_window_width()
        for key in sorted(
            group_by_window_width.keys()
        ):  # looping over group_by_window_width.items() gives wrong order
            value: pd.DataFrame = deepcopy(group_by_window_width[key])
            value.sort_values(
                by=self.window_start_col_name,
                ascending=True,
                ignore_index=True,
                inplace=True,
            )
            x_arrays.append(value.loc[:, self.window_start_col_name])
            y_arrays.append(value.loc[:, col_name])
            labels.append(str(key))  # noqa: key is a string

        # Final settings and overrides
        if legend_override:
            labels: List[str] = legend_override
        kwargs.setdefault("types", ["scatter", "plot"])
        kwargs.setdefault("xlabel", self.translate(self.translate("window_start")))
        kwargs.setdefault("ylabel", self.translate(col_name))
        kwargs.setdefault("title", self.translate(self.name_root))
        return visualize_x_y(
            x_arrays,
            y_arrays,
            legend_strings=labels,
            style=Style(**self.dict()),
            **kwargs,
        )

    def plot_pdfs(self, col_name: str, legend_override: List[str] = None, **kwargs) -> plt.Figure:
        """
        Plot the probability density function(s) of the sliding window results. Useful kwargs: [cumulative, density]

        :param col_name: which field should be plotted
        :param legend_override: custom legend labels if desired (must have same length as number of unique widths)
        :param kwargs: plt.hist() keyword arguments.
        :return: figure handle
        """
        grouped: Dict[float, pd.DataFrame] = self.group_by_window_width()
        data_sets: List[Any] = []
        labels: List[Any] = []
        for key, value in grouped.items():
            data_sets.append(value[col_name])
            labels.append(str(key))
        if legend_override:
            labels: List[str] = legend_override
        kwargs.setdefault("xlabel", self.translate(col_name))
        kwargs.setdefault("title", self.translate(self.name_root))
        return visualize_1d_distribution(data_sets, legend_strings=labels, **kwargs)

    def heatmap_feature(self, col_name: str, **kwargs) -> plt.Figure:
        """
        Returns a heatmap of some feature as a function of window start time and start width

        :param col_name: which column to plot
        :param kwargs: additional keyword arguments for surface method
        :return: figure handle of the plot
        """
        return self.surface(self.window_start_col_name, self.window_width_col_name, col_name, **kwargs)


class InfoSurface(SlidingWindowResults):
    """Wrapper for SlidingWindowResults that knows how to plot the infosurface"""

    def plot_info_surface(self, singular_values: List[int] = None, **kwargs) -> List[plt.Figure]:
        """Plot the info surface showing value of singular vectors as a function of window start and width

        :param singular_values: Which singular values to plot [1, 2, 3] by default
        :return: figure handles
        """
        figure_handles: List[plt.Figure] = []
        if not singular_values:
            singular_values: List[int] = [1, 2, 3]
        for s in singular_values:
            figure_handles.append(
                self.heatmap_feature(f"singular_value_{s}", title=f"Singular value # {s}", **kwargs)
            )
        return figure_handles


class VectorSequence(VectorMultiset):
    """Set of vectors in a sequence ordered according to some basis (e.g. height, wavelength, time)"""

    # (Inherits `data` dataframe and `name_root` string from VectorMultiset)
    basis_col_name: str = "basis"
    error_if_basis_quality_issues: bool = False
    info_surface: InfoSurface = None

    def __init__(self, skip_vector_sequence_init: bool = False, **kwargs: Any):
        super().__init__(**kwargs)

        # Cast to numeric if possible
        if not kwargs.get("disable_auto_conversion_to_numeric", False) and (self.data is not None):
            self.cast_to_numeric()

        # Handle the other processing specific to ordered data
        if (self.data is not None) and len(self.data) and (not skip_vector_sequence_init):
            if (not self.basis_col_name) or (self.basis_col_name not in self.data.keys()):
                raise ValueError(
                    f"{self.basis_col_name=} not in known keys: {self.data.keys().tolist()=}\n"
                    + "Pass skip_vector_sequence_init=True to suppress"
                )
            self.sort(by=self.basis_col_name, inplace=True, reset_index=True)
            if self.error_if_basis_quality_issues:
                is_ok, explanation = self.passes_basis_quality_checks()
                if not is_ok:
                    raise ValueError(
                        f"Issues with basis:\n{explanation}\nSee: `error_if_basis_quality_issues`"
                    )

    def basis(self) -> List[Any]:
        """Thin wrapper that returns the basis as a list"""
        return self.data.loc[:, self.basis_col_name].tolist()

    def basis_quality_checks(self) -> Tuple[bool, str]:
        """Checks basis data quality and returns both a True/False flag and a string with an explanation

        :return: True if OK, and a string with explanation either way
        """
        return basis_quality_checks(self.basis())

    def passes_basis_quality_checks(self) -> bool:
        """Super thin wrapper around basis_quality_checks that drops the explanation string

        :return: True if OK
        """
        result, _ = self.basis_quality_checks()
        return result

    def basis_quality_plots(self, **kwargs) -> List[plt.Figure]:
        """Creates a series of plots showing the basis data quality (missing data, nans, timeseries gaps, etc)

        :return: figure handles for several plots
        """
        return basis_quality_plots(self.basis(), **kwargs)

    def fill_ratio(self) -> float:
        """Checks for how many data points are observed for the expected number of data points in a uniform basis.
            Attempts to infer spacing from median of diffs. If used on non-uniform data the result is not meaningful.

        Use example: [10, 20, 40, 50] has 0.8 fill ratio because it is missing the 30 to complete the set

        :return: fractional amount of the array
        """
        return fill_ratio(self.data.loc[:, self.basis_col_name])

    def slice(
        self,
        start_at: Any = None,
        stop_at: Any = None,
        inplace: bool = False,
        reset_index: bool = True,
        return_type: str = None,
        **kwargs,
    ) -> Union[None, pd.DataFrame, Tuple[Iterable, Iterable], Any]:
        """Slices the VectorSequence according to the basis

        # TODO have slice return self type for convenience

        :param start_at: Start point (can be None)
        :param stop_at: Stop point (can be None)
        :param inplace: slice inplace or return the result (default --> return)
        :param return_type: how to return the data ('VectorMultiset', 'VectorSequence', 'dataframe', etc)
        :param reset_index: indicate whether or not to reset the data frame index (default --> True)
        :param kwargs: keyword arguments passed through on init of non-dataframe objects
        :return: VectorSequence if not inplace
        """
        df: pd.DataFrame = deepcopy(self.data)
        df.sort_values(by=self.basis_col_name, ascending=True, inplace=True, ignore_index=True)

        # If inputs are strings, attempt to convert to timestamps
        if isinstance(start_at, str):
            try:
                start_at: float = machine_time(start_at)
            except ValueError:
                raise ValueError(f"Unsure how to process {start_at=}")
        if isinstance(stop_at, str):
            try:
                stop_at: float = machine_time(stop_at)
            except ValueError:
                raise ValueError(f"Unsure how to process {stop_at=}")

        if not start_at:
            start_at: Any = min(df.loc[:, self.basis_col_name])
        if not stop_at:
            stop_at: Any = max(df.loc[:, self.basis_col_name])
        in_range: pd.DataFrame = df.loc[df.loc[:, self.basis_col_name].between(start_at, stop_at), :]
        if reset_index:
            in_range.reset_index(drop=True, inplace=True)
        if inplace:
            self.data = in_range
        else:
            if not return_type:
                return self.__class__(
                    basis_col_name=self.basis_col_name,
                    name_root=self.name_root,
                    data=in_range,
                    **kwargs,
                )
            else:
                if "sequence" in (return_type_lower := return_type.lower()):
                    return VectorSequence(
                        basis_col_name=self.basis_col_name,
                        name_root=self.name_root,
                        data=in_range,
                        **kwargs,
                    )
                elif "dataframe" in return_type_lower:
                    return in_range
                elif "timeseries" in return_type_lower:
                    return Timeseries(
                        basis_col_name=self.basis_col_name,
                        name_root=self.name_root,
                        data=in_range,
                        **kwargs,
                    )
                else:
                    raise ValueError(f"Unknown return type: {return_type}")

    def split(self, split_at: Union[float, str], gap: float = 0, **kwargs) -> tuple:
        """
        Splits the sequence into two subsequences (with an optional gap, which is placed after sthe split point)

        :param split_at: basis value to split
        :param gap: whether to leave a gap between the end of the pre-split and the start of the post-split (optional)
        :param kwargs: additional keyword arguments passed to the slice method (e.g. reset_index, return_type)
        :return: two objects separate at the split point
        """
        kwargs["inplace"] = False  # Cannot split one object in place
        pre_split = self.slice(stop_at=split_at, **kwargs)
        post_split = self.slice(start_at=split_at + gap, **kwargs)
        return pre_split, post_split

    def index_to_basis(self) -> None:
        """Adds (or overwrites) 'basis' column with a default index [0, 1, 2, ...]"""
        self.data.loc[:, self.basis_col_name] = list(range(len(self.data)))

    def sort(self, by: str = None, inplace: bool = True, reset_index: bool = True):
        """Sorts the data points by the basis

        :param inplace: import inplace (default) or return the result
        :param reset_index: indicate whether or not to reset the data frame index (default = True)
        :param by: which field to use for the sort (uses the basis by default)
        :return: sorted data frame, if inplace is False
        """
        if not by:
            by = self.basis_col_name
        result: pd.DataFrame = self.data.sort_values(by=by, ascending=True, inplace=False, ignore_index=False)
        if reset_index:
            result.reset_index(drop=True, inplace=True)
        if inplace:
            self.data = result
        else:
            return VectorSequence(
                basis_col_name=self.basis_col_name,
                name_root=self.name_root,
                data=result,
            )

    # Resampling function (same overall shape as the downsampling function, except with only the pandas engine functionality
    def resample(
        self,
        interval: Union[int, float],
        inplace: bool = True,
        target_basis_vector: Union[List[float], None] = None,
        direction: str = "backward",
        **kwargs,
    ):
        """
        Resamples a VectorSequence

        :param interval: interval by which to resample (either by N rows, or by some basis interval)
        :param inplace: whether to update the dataframe data attribute in place or return an updated copy of self
        :param target_basis_vector: optionally, specify an array of timestamps to resample to (default --> None)
        :param direction: whether to resample forward or backward (default --> "backward" to avoid data leaks!)
        :param kwargs: additional keyword arguments passed to the resample method (e.g. reset_index, return_type)
        :return: resampled VectorSequence
        """
        if target_basis_vector is not None:
            target_timebase: pd.DataFrame = pd.DataFrame({self.basis_col_name: target_basis_vector})
        else:
            basis_df = pd.DataFrame({self.basis_col_name: self.basis()})
            min_basis = basis_df[self.basis_col_name].min()
            max_basis = basis_df[self.basis_col_name].max()
            target_timebase: pd.DataFrame = pd.DataFrame(
                {self.basis_col_name: np.arange(min_basis, max_basis, interval)}
            )
        df_merged: pd.DataFrame = pd.merge_asof(
            target_timebase,
            self.data,
            left_on=self.basis_col_name,
            right_on=self.basis_col_name,
            direction=direction,
            **kwargs,
        )
        if inplace:
            self.data = df_merged
        else:
            return df_merged

    # Downsampling function
    def downsample(
        self,
        interval: Union[int, float],
        method: str = "by_basis",
        basis_downsample_engine: str = "legacy",
        inplace: bool = True,
        target_timestamps: Union[List[float], None] = None,
        suppress_deprecation_warning: bool = False,
        **kwargs,
    ):
        """
        Downsamples a VectorSequence

        If downsampling by basis, legacy mode leaves gaps where there is no data, whereas the pandas engine returns a regular basis

        :param interval: interval by which to downsample (either by N rows, or by some basis interval)
        :param method: 'by_basis' or 'by_row'
        :param inplace: whether to update the dataframe data attribute in place or return an updated copy of self
        :param basis_downsample_engine: How to downsample based on time 'pandas' or 'legacy' (default --> 'pandas')
        :param target_timestamps: optionally, specify an array of timestamps to downsample to (default --> None)
        :param suppress_deprecation_warning: whether to suppress warnings (default --> False)
        :param kwargs: additional keyword arguments passed to the merge_asof method from padas
        :return: Downsampled copy of self (or nothing if 'inplace=True')
        """
        result_vector: Any = deepcopy(self)
        result_vector.sort(inplace=True, reset_index=True)

        if "by_row" in method.lower():
            if not isinstance(interval, int):
                raise ValueError(f"downsample with method='by_row' requires integer interval")
            result_vector.data = result_vector.data.iloc[::interval, :]

        elif "by_basis" in method:
            if basis_downsample_engine.lower() == "legacy":
                keep_indices: List[int] = []
                last: float = result_vector.basis()[0]
                for i, basis_value in enumerate(self.basis()):
                    if basis_value >= last + interval:
                        keep_indices.append(i)
                        last = basis_value
                result_vector.data = result_vector.data.iloc[keep_indices, :]
            elif basis_downsample_engine.lower() == "pandas":
                if not suppress_deprecation_warning:
                    print(
                        "... Please upgrade your code to use .resample() instead of .downsample(), at your leisure"
                        "\n you can suppress this warning by passing suppress_deprecation_warning=True"
                    )
                if target_timestamps is not None:
                    target_timebase: pd.DataFrame = pd.DataFrame({self.basis_col_name: target_timestamps})
                else:
                    basis_df = pd.DataFrame({self.basis_col_name: self.basis()})
                    min_basis = basis_df[self.basis_col_name].min()
                    max_basis = basis_df[self.basis_col_name].max()
                    target_timebase: pd.DataFrame = pd.DataFrame(
                        {self.basis_col_name: np.arange(min_basis, max_basis, interval)}
                    )
                result_vector.data = pd.merge_asof(
                    target_timebase,
                    result_vector.data,
                    left_on=self.basis_col_name,
                    right_on=self.basis_col_name,
                    direction="backward",
                    **kwargs,
                )
            else:
                raise ValueError(
                    f"Unknown basis_downsample_engine: {basis_downsample_engine}. Use 'legacy' or 'pandas'"
                )

        else:
            raise ValueError(f"Unknown downsample method {method}; try 'by_row' or 'by_basis'.")

        # Downsample and record or return the result
        if inplace:
            self.data = result_vector.data
        else:
            return result_vector

    def evaluate_over_window(
        self,
        function: Callable,
        start_at: float,
        window_width: float,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Helper function that evaluates a function over a given window

        :param function: function to be applied to the window
        :param start_at: start time for the window
        :param window_width: width of the window
        :param args: positional arguments for the function (NB: args, not *args)
        :param kwargs: keyword arguments for the function (NB: kwargs, not **kwargs)
        :return: The evaluation of the function (along with a note of the window start and width)
        """
        window_data: VectorSequence = self.slice(
            start_at=start_at,
            stop_at=start_at + window_width,
            inplace=False,
            reset_index=True,
        )
        try:
            return {
                **function(window_data, *args, **kwargs),
                "window_start": start_at,
                "window_width": window_width,
            }
        except IndexError as e:
            window_description: str = f"{start_at=}\n{window_width=}\n"
            if not len(window_data):
                raise IndexError(
                    "\nCaught IndexError.\n\nHint: there were no data points in this window:\n"
                    + f"{window_description}\n(The best way to resolve this is probably to tweak\n"
                    + f"your eval function to gracefully handle empty slices)\n\nOriginal error:\n{e}"
                )
            else:
                raise IndexError(
                    f"Eval function had IndexError at window:\n{window_description}\nOriginal error:\n{e}"
                )
        except TypeError as e:
            raise TypeError(
                f"Caught type error (below). Hint: your eval function needs to return a dictionary.\n\n{e}"
            )

    #                                       vv sequence vv       vv args vv          vv kwargs vv
    def sliding_window(
        self,
        function: Callable[[Any, Union[Tuple[Any], List[Any]], Dict[str, Any]], Dict[str, Any]],
        window_widths: List[float] = None,
        window_starts: List[Any] = None,
        step_size: float = None,
        parallelize_sliding_window: Union[bool, int] = True,
        allow_shorter_windows: bool = False,
        disable_sliding_window_progress_bar=None,
        limit: int = None,
        *args,
        **kwargs,
    ) -> SlidingWindowResults:
        """Apply function in a sliding window over the sequence

        :param function: callable to be applied
        :param window_widths: list of window widths to use (or a single value if only one width desired)
        :param window_starts: list of starting points for the windows
        :param step_size: how far apart to space windows
        :param allow_shorter_windows: whether to allow windows that are not the full length (disabled by default)
        :param args: positional arguments for the function
        :param kwargs: keyword arguments for the function
        :param parallelize_sliding_window: whether to use multiprocessing for the sliding window
        :param disable_sliding_window_progress_bar: pass True to disable progress bar (only applies to serial mode)
        :param limit: set a limit for how many windows to evaluate, takes the first N
        :return: SlidingWindowResults (see the dataframe attribute for results)
        """

        # Validation
        if not self.basis_col_name:
            raise AttributeError(
                "cannot apply sliding window analysis without knowing which field is the basis"
            )
        if self.basis_col_name not in self.data.keys():
            raise AttributeError(
                f"basis_col_name {self.basis_col_name} not a column in data frame:\n{self.data}"
            )

        basis: Tuple[float] = self.data.loc[:, self.basis_col_name].tolist()

        if isinstance(window_widths, (float, int)):
            window_widths = [window_widths]

        if not window_widths:
            duration: float = max(self.data.loc[:, self.basis_col_name]) - min(
                self.data.loc[:, self.basis_col_name]
            )
            window_widths: List[float] = [duration / x for x in range(20, 401, 20)]  # TODO: move vals to vars

        # First, figure out where to place the windows based on the window widths and/or steps size:
        list_of_start_and_width_tuples: List[Tuple[float, float]] = []
        for i, window_width in enumerate(window_widths):
            if not window_starts:
                if step_size:
                    if allow_shorter_windows:
                        window_starts: List[float] = list(np.arange(min(basis), max(basis), step_size))
                    else:
                        window_starts: List[float] = list(
                            np.arange(min(basis), max(basis) - window_width, step_size)
                        )
                else:
                    window_starts: List[float] = list(
                        np.arange(min(basis), max(basis) - window_width, window_width)
                    )
            list_of_start_and_width_tuples += [(start_time, window_width) for start_time in window_starts]

        if not list_of_start_and_width_tuples:
            raise ValueError("No windows for sliding analysis - check your sizes and start times")

        if limit and limit < len(list_of_start_and_width_tuples):
            list_of_start_and_width_tuples = list_of_start_and_width_tuples[:limit]

        # If parallelize_sliding_window != False, run in parallel using starmap() from multiprocessing library `Pool`
        num_workers: int = get_num_workers(parallelize_sliding_window)
        if parallelize_sliding_window and (num_workers > 1):
            evaluations: List[Dict[Any, Any]] = process_queue(
                func=self.evaluate_over_window,
                iterable=[
                    (function, start, width, args, kwargs) for start, width in list_of_start_and_width_tuples
                ],
                pool_function="starmap",
                batching=False,
                num_workers=num_workers,
            )

        # If parallelize_sliding_window == False (or None) use `for` loop in list comprehension -> serial processing
        else:
            evaluations: List[Dict[Any, Any]] = [
                self.evaluate_over_window(
                    function=function,
                    start_at=start,
                    window_width=width,
                    args=args,
                    kwargs=kwargs,
                )
                for start, width in tqdm(
                    list_of_start_and_width_tuples,
                    disable=disable_sliding_window_progress_bar,
                )
            ]

        return SlidingWindowResults(
            window_width_col_name="window_width",
            window_start_col_name="window_start",
            data=pd.DataFrame(evaluations),
            name_root=self.name_root,
        )

    def repackage(
        self,
        instance: Any,
        sequence_attribute: str = "series",
        basis_name: str = "basis",
    ) -> Any:
        """
        Helper function for repackaging similar (BaseModel-like) objects into the top class

        :param instance: thing to be repackaged
        :param sequence_attribute: the name of the sequence (read: values) attribute
        :param basis_name: the name of the basis (read: time-like) attribute
        :return: repackaged instance with new class (if applicable)
        """
        cache: Dict[str, Any] = instance.dict()
        cache["data"] = pd.DataFrame({x: input.__getattribute__(x) for x in [sequence_attribute, basis_name]})
        cache["basis_col_name"] = basis_name
        return self.__class__(**cache)

    def seasonal_decompose(self, col: str, period: float, **kwargs):
        """Seasonal (weekly, monthly, etc) decomposition, must specify the period

        :param period: period of the data
        :param col: which data feature to use
        :param kwargs: additional kwargs for statsmodels.tsa.seasonal_decompose
        :return: sm.tsa.seasonal.DecomposeResult
        """
        x: pd.DataFrame = deepcopy(self.data.loc[:, col])
        try:
            return sm.tsa.seasonal_decompose(x, period=period, **kwargs)
        except ValueError as e:
            raise ValueError(f"Error - did you include a non-numeric column, or not specify 'cols'? See: {e}")

    def plot_decomposition(
        self,
        col: str,
        period: float,
        figsize: Tuple[float, float] = None,
        which_plots: List[str] = None,
        xlabel: str = "basis",
        ylabel: str = "[units]",
        title: str = None,
        xlim: List[float] = None,
        **kwargs,
    ) -> List[plt.Figure]:

        """Plot the seasonal (weekly, monthly, etc) decomposition. Must specify period

        :param period: period of the data
        :param col: which data feature to use
        :param figsize: size of each subfigure
        :param which_plots: which plots to make
        :param xlabel: how to label the x-axis
        :param ylabel: how to label the y-axis (e.g. units)
        :param title: title string (prefix) for the plots
        :param xlim: x-axis limits (Left, Right)
        :param kwargs: additional kwargs for statsmodels.tsa.seasonal_decompose
        :return: list of figure handles
        """
        if not figsize:
            figsize: List[float] = [8.0, 3.0]
        if not which_plots:
            which_plots: List[str] = ["observed", "trend", "seasonal", "residual"]
        if not title:
            title: str = self.translate(self.name_root)

        def add_labels(x_label: str, y_label: str) -> None:
            plt.xlabel(x_label, size=self.label_fontsize)
            plt.ylabel(y_label, size=self.label_fontsize)
            plt.legend(fontsize=self.legend_fontsize)

        decomposition = self.seasonal_decompose(col, period, **kwargs)
        figure_handles: List[plt.Figure] = []

        if "observed" in which_plots:
            figure_handles.append(plt.figure(figsize=figsize, facecolor=self.facecolor))
            plt.plot(
                self.data.loc[:, self.basis_col_name],
                decomposition.observed,
                "-",
                label="Sequence",
                color="black",
            )
            add_labels(xlabel, ylabel)
            plt.title(f"{title}observed", size=self.title_fontsize)
            if xlim:
                plt.xlim(xlim)

        if "trend" in which_plots:
            figure_handles.append(plt.figure(figsize=figsize, facecolor=self.facecolor))
            plt.plot(
                self.data.loc[:, self.basis_col_name],
                decomposition.trend,
                "-",
                label="Trend",
                color="green",
            )
            add_labels(xlabel, ylabel)
            plt.title(f"{title}trend", size=self.title_fontsize)
            if xlim:
                plt.xlim(xlim)

        if "seasonal" in which_plots:
            figure_handles.append(plt.figure(figsize=figsize, facecolor=self.facecolor))
            plt.plot(
                self.data.loc[:, self.basis_col_name],
                decomposition.seasonal,
                "-",
                label="Seasonality",
                color="darkslateblue",
            )
            add_labels(xlabel, ylabel)
            plt.title(f"{title}seasonality with period {period}", size=self.title_fontsize)
            if xlim:
                plt.xlim(xlim)

        if "residual" in which_plots:
            figure_handles.append(plt.figure(figsize=figsize, facecolor=self.facecolor))
            plt.plot(
                self.data.loc[:, self.basis_col_name],
                decomposition.resid,
                "-",
                label="Residual",
                color="darkred",
            )
            add_labels(xlabel, ylabel)
            plt.title(f"{title}residual", size=self.title_fontsize)
            if xlim:
                plt.xlim(xlim)

        return figure_handles

    def singular_value_decomposition(
        self, cols: Union[str, List[str]] = None, cache_results: bool = False, **kwargs
    ) -> SVD:
        """Applies singular value decomposition, skipping the basis column

        :param cache_results: if True, saves u, s, and vh as the attribute 'svd' (an SVD class)
        :param cols: which data features to use
        :param kwargs: keyword arguments for SVD
        :return: u, s, vh arrays
        """
        if not cols:
            cols = [x for x in self.data.keys().tolist() if x != self.basis_col_name]
        svd: SVD = singular_value_decomposition(self.data, cols, **kwargs)
        if cache_results:
            self.svd = svd
        return svd

    def calculate_info_surface(
        self,
        window_widths: List[float] = None,
        cols: Union[str, List[str]] = None,
        *args,
        save_result: bool = True,
        **kwargs,
    ) -> InfoSurface:
        """Calculates the info surface by sliding the SVD function along the basis

        :param window_widths: window widths for the sliding window analysis
        :param cols: which data features to use in the svd
        :param args: args for sliding window analysis and eval function
        :param save_result: whether to save attach the infosurface to the object once calculated
        :param kwargs: kwargs for sliding window analysis and eval function
        :return: InfoSurface object
        """
        result: SlidingWindowResults = self.sliding_window(
            info_surface_slider, window_widths=window_widths, cols=cols, *args, **kwargs
        )
        info_surface: InfoSurface = InfoSurface(**result.dict())
        if save_result:
            self.info_surface = info_surface
        return info_surface

    def plot_info_surface(
        self,
        window_widths: List[float] = None,
        cols: Union[str, List[str]] = None,
        singular_values: List[int] = None,
        style: Style = None,
        use_cache: bool = False,
        *args,
        **kwargs,
    ) -> List[plt.Figure]:
        """Calculates and plots the info surface: value of singular vectors as a function of window start and width

        :param window_widths: window widths for the sliding window analysis
        :param cols: which data features to use in the svd
        :param singular_values: singular values for which to plot the surface (default: [1, 2, 3])
        :param use_cache: uses cached info surface if available, otherwise calculates a fresh one
        :param args: args for sliding window analysis and eval function
        :param kwargs: kwargs for sliding window analysis and eval function
        :param style: isthmuslib Style object for the colormap
        :return: list of figure handles
        """
        # Set style. Overrides: kwargs > style input > Style() defaults

        config: Style = Style(**{**Style().dict(), **make_dict(style), **make_dict(kwargs)})
        svd_kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if k not in config.dict()}
        style_kwargs: Dict[str, Any] = {k: v for k, v in kwargs.items() if k in config.dict()}

        if use_cache and self.info_surface:
            info_surface: InfoSurface = self.info_surface
        else:
            info_surface: InfoSurface = self.calculate_info_surface(
                window_widths=window_widths, cols=cols, *args, **svd_kwargs
            )
        return info_surface.plot_info_surface(
            singular_values=singular_values,
            **{**make_dict(style), **make_dict(style_kwargs)},
        )

    def correlation_matrix(self, exclude_basis: bool = True, **kwargs) -> pd.DataFrame:
        """
        Very thin wrapper around correlation matrix (using pandas corr() method)

        :param exclude_basis: whether to exclude the basis column from correlation analysis
        :param kwargs: additional keyword arguments for correlation_matrix (fed through to df.style.background_gradient)
        :return: styled pandas dataframe
        """
        if exclude_basis:
            if kwargs_in := kwargs.get("exclude_cols"):
                if self.basis_col_name not in kwargs_in:
                    kwargs["exclude_cols"] = kwargs_in + [self.basis_col_name]
            else:
                kwargs: Dict[str, Any] = {"exclude_cols": self.basis_col_name}
        return correlation_matrix(self.data, **kwargs)

    def downsample_and_plot_matrix_profile_univariate(
        self,
        downsample_interval: float,
        col_names: Union[str, Iterable[str]],
        diff: bool = False,
        downsample_method: str = "by_basis",
        **kwargs,
    ) -> List[Tuple[Any, List[plt.Figure]]]:
        """
        Downsamples and optionally diffs data before applying the matrix profile (to which kwargs are passed through)

        :param downsample_interval: interval by which to downsample (either by N rows, or by some basis interval)
        :param downsample_method: 'by_basis' or 'by_row'
        :param col_names: a column name (or list of column names) to be profiled
        :param diff: whether to take the diff of the rows (coming soon - fractional differentiation??)
        :param kwargs: keyword arguments passed through to matrixprofile.analyze()
        :return: List of outputs from matrixprofile.analyze()
        """
        downsampled = self.downsample(
            downsample_interval, method=downsample_method, inplace=False
        )  # self type
        if diff:
            downsampled.data = downsampled.data.diff()
            downsampled.data.dropna(inplace=True)
        downsampled.data.reset_index(inplace=True)
        return downsampled.matrix_profile_univariate(col_names, **kwargs)

    # def plot_fluss_semantic_segmentation(
    #     self,
    #     col_name: str = None,
    #     fluss_width: int = 15,
    #     n_regimes: int = 2,
    #     excl_factor: float = 1,
    #     **kwargs,
    # ) -> plt.Figure:
    #     """
    #     Wrapper for STUMPY's FLUSS algonithm + plots
    #
    #     :param col_name: name of the column to plot
    #     :param fluss_width: FLUSS window size
    #     :param n_regimes: number of regimes in the data
    #     :param excl_factor: factor to correct for the start and end of the arc curve
    #     :param kwargs: additional keyword arguments for stumpy and plots (figsize, title, ...)
    #     :return: figure handle showing the data and the analysis results
    #     """
    #     figsize: Any = kwargs.pop("figsize", self.figsize)  # typically a 2-element tuple or list
    #     title: str = kwargs.pop("title", self.translate(self.name_root))
    #
    #     # If column name not specified and only one non-basis column, infer that we should use that one
    #     if not col_name:
    #         if len(self.data.keys()) == 2:
    #             col_name: str = [x for x in self.data.keys() if x != self.basis_col_name][0]
    #         else:
    #             raise ValueError(
    #                 f"For fluss, please specify `col_name=X` with X from: {list(self.data.keys())}"
    #             )
    #
    #     # Calculate the matrix profile and FLUSS results
    #     data_array: np.ndarray = self.data.loc[:, col_name].to_numpy()
    #     mp: np.ndarray = stumpy.stump(data_array, m=fluss_width, **kwargs)
    #     cac, regime_locations = stumpy.fluss(
    #         mp[:, 1], L=fluss_width, n_regimes=n_regimes, excl_factor=excl_factor
    #     )
    #
    #     # Plot the results
    #     f, axs = plt.subplots(
    #         2,
    #         sharex=True,  # noqa: bool is in fact allowed
    #         gridspec_kw={"hspace": 0},
    #         figsize=figsize,
    #         facecolor=self.facecolor,
    #     )
    #     plt.suptitle(title, fontsize=self.title_fontsize)
    #     axs[0].plot(range(data_array.shape[0]), data_array, color=self.color)
    #     axs[0].axvline(x=regime_locations[0], linestyle="dashed")
    #     axs[1].plot(range(cac.shape[0]), cac, color="k")
    #     axs[1].axvline(x=regime_locations[0], linestyle="dashed")
    #     return f

    def human_time_start_and_stop(self, **kwargs) -> Tuple[str, str]:
        """
        Helper function that returns the min and max value of the basis column converted to human-readable string

        :param kwargs: keyword arguments for human_time()
        :return: tuple of strings like ('2025-03-21', '2028-07-04')
        """
        return self.human_time(
            min(self.data.loc[:, self.basis_col_name].astype(float).dropna().tolist()),
            **kwargs,
        ), self.human_time(
            max(self.data.loc[:, self.basis_col_name].astype(float).dropna().tolist()),
            **kwargs,
        )

    def human_timeframe(self, prefix: str = None, between: str = None, suffix: str = None, **kwargs) -> str:
        """
        Formatting helper function for ensuring consistency across timeframe descriptions, for example

            ``From 2025-03-21 to 2028-07-04``

        :param prefix: text to precede a time/date range
        :param between: text to go between the times/dates
        :param suffix: text to follow the time/date range
        :param kwargs: additional keyword arguments for human_time, e.g. formatter = '%Y-%m-%d %H:%M:%S'
        :return: a single string describing the timeframe
        """
        kwargs.setdefault("formatter", self.formatter)
        kwargs.setdefault("include_timezone", self.include_timezone)
        if prefix is None:
            prefix: str = self.timeframe_prefix
        if between is None:
            between: str = self.timeframe_between
        if suffix is None:
            suffix: str = self.timeframe_suffix
        start_time_string, stop_time_string = self.human_time_start_and_stop(**kwargs)
        return f"{prefix}{start_time_string}{between}{stop_time_string}{suffix}"


class Timeseries(VectorSequence):
    """Thin wrapper for VectorSequence in the context of time"""

    basis_col_name: str = "timestamp"

    # x_axis_human_tick_labels: bool = True

    def calc_weights(
        self,
        lowest_weight: float = 0.5,
        decay_timeframe_sec: float = None,
        show_plots: bool = False,
        method: str = "linear",
        post_window_weight: float = None,
        base: float = np.e,
    ) -> List[float]:
        """
        Calculates a vector of weights from an exponential or linear decay

        :param lowest_weight: what should be the minimum weight? (i.e. at the end of the decay timeframe)
        :param decay_timeframe_sec: how long should the decay window be?
        :param post_window_weight: what should weights be after the decay window? Set to 0 to forget anything older.
        :param show_plots: whether to show plots of the weights
        :param method: whether the decay should be linear or exponential
        :param base: the base for the decay, but I don't think it matters
        :return: a list of weights
        """
        min_timestamp: float = min(self.data.loc[:, self.basis_col_name])
        max_timestamp: float = max(self.data.loc[:, self.basis_col_name])

        # If decay_timeframe_sec not specified, apply to the whole data set
        if decay_timeframe_sec is None:
            decay_timeframe_sec = max_timestamp - min_timestamp

        # If post_window_weight not specified, carry the last weight forward
        if post_window_weight is None:
            post_window_weight = lowest_weight

        # Make sure that we don't collide on column name
        age_col_name: str = "age"
        i: int = 0
        while age_col_name in self.data.keys():
            age_col_name += "_"
            i += 1
            if (
                i > 10
            ):  # this 10 is not a magic number, I just picked age_, age__, age___, up to 10 as excessive
                raise NotImplementedError(
                    f"Having {i} age columns with trailing underscores is not allowed..."
                )

        # Calculate age in a copy of the data frame
        df: pd.DataFrame = deepcopy(self.data)
        df[age_col_name] = max_timestamp - df.loc[:, self.basis_col_name]

        # Calculate the weights (not taking into account the decay timeframe)
        if "lin" in method.lower():  # 'lin' or 'linear'
            #  y = m * x + b
            slope: float = (lowest_weight - 1) / decay_timeframe_sec
            weights_all: List[float] = [slope * x + 1 for x in df.loc[:, age_col_name]]
        elif "exp" in method.lower():  # 'exp' or 'exponential'
            #  y = b ^ (-k * x) where k is the decay constant and x is the age
            decay_constant: float = -1 * math.log(lowest_weight, base) / decay_timeframe_sec
            weights_all: List[float] = [
                base ** (-1 * decay_constant * tau) for tau in df.loc[:, age_col_name]
            ]
        else:
            raise ValueError(f"Unknown method {method}, please try 'linear' or 'exponential'")

        # Replace any out-of-timeframe data points with the post_window_weight
        ages_and_weights: zip = zip(df.loc[:, age_col_name], weights_all)
        weights_raw: List[float] = [
            w if a < decay_timeframe_sec else post_window_weight for a, w in ages_and_weights
        ]
        normalization_factor = sum(weights_raw)
        final_weights: List[float] = [w / normalization_factor for w in weights_raw]

        # Show plots of weights for the curious user
        if show_plots:
            self.scatter(
                df.loc[:, self.basis_col_name].tolist(),
                final_weights,
                xlabel=self.translate(self.basis_col_name),
                ylabel="weights",
                title="weights vs timestamps",
            )

            self.scatter(
                df.loc[:, age_col_name].tolist(),
                final_weights,
                xlabel="age",
                ylabel="weights",
                title="weights vs ages",
            )
            plt.show()

        return final_weights

    def weighted_mean(self, col_name: str, **kwargs) -> float:
        """
        Calculates the weighted mean of the provided column name. kwargs are passed to calc_weights

        :param col_name: which column to calculate the weight for
        :param kwargs: kwargs for calc_weights, e.g. lowest_weight, decay_timeframe_sec, post_window_weight, method
        :return: weighted mean of that column
        """

        weights = self.calc_weights(**kwargs)
        return sum([weight * val for weight, val in zip(weights, self.data.loc[:, col_name])])


################
# Helpers
################


def correlation_matrix(
    dataframe: pd.DataFrame,
    use_cols: List[str] = None,
    exclude_cols: List[str] = None,
    correlation_method: str = "pearson",
    style: Style = None,
    display_row_names: Optional[List[str]] = None,
    display_col_names: Optional[List[str]] = None,
    auto_filter_to_numeric: bool = True,
    min_periods: Optional[int] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Returns a styled pandas data frame with correlation coefficients (Pearson by default), wraps pandas.corr()

    :param dataframe: dataframe to analyze
    :param use_cols: which columns to use
    :param exclude_cols: which columns to exclude (NB: exclude overrides include)
    :param correlation_method: Method of correlation {‘pearson’, ‘kendall’, ‘spearman’} or Callable
    :param style: isthmuslib Style object for the colormap
    :param display_row_names: which row names to display in the final output
    :param display_col_names: which column names to display in the final output
    :param auto_filter_to_numeric: automatically filter non-numeric columns
    :param kwargs: additional keyword arguments for df.style.background_gradient()
    :return: styled pandas dataframe
    """
    if not use_cols:
        use_cols: List[str] = dataframe.keys().tolist()
    if exclude_cols:
        use_cols: List[str] = [x for x in use_cols if x not in exclude_cols]

    # Filter to numeric columns if auto_filter_to_numeric is True
    if auto_filter_to_numeric:
        dataframe = dataframe.select_dtypes(include=np.number)
        use_cols: List[str] = [col for col in use_cols if col in dataframe.columns]

    if style:
        kwargs.setdefault("cmap", style.cmap)
    else:
        kwargs.setdefault("cmap", Style().cmap)

    corr: pd.DataFrame = dataframe.loc[:, use_cols].corr(method=correlation_method, min_periods=min_periods)

    # Filter the correlation matrix by the provided row and column names, if they are not None
    if display_row_names is not None:
        display_row_names = [row for row in display_row_names if row in corr.index]
        corr = corr.loc[display_row_names, :]
    if display_col_names is not None:
        display_col_names = [col for col in display_col_names if col in corr.columns]
        corr = corr.loc[:, display_col_names]

    return corr.style.background_gradient(**kwargs)  # type: ignore


def info_surface_slider(vs: VectorSequence, *args, **kwargs) -> Dict[str, Any]:
    """
     Helper function for sliding window SVD analysis

    :param vs: vector sequence
    :param args: args for svd
    :param kwargs: kwargs for svd
    :return: dictionary with singular values
    """
    svd: SVD = vs.singular_value_decomposition(*args, **kwargs)
    return {f"singular_value_{i + 1}": value for i, value in enumerate(svd.s)}


def singular_value_decomposition(
    df: pd.DataFrame,
    cols: Union[str, List[str]] = None,
    normalize: bool = False,
    **kwargs,
) -> SVD:
    """
    Helper function that wraps numpy svd to feed in a subset of data features (see kwarg: 'full_matrices')

    :param df: pandas dataframe to analyze
    :param cols: which data features to use
    :param normalize: whether to normalize the data to zero mean and unit variance
    :param kwargs: keyword arguments for SVD
    :return: u, s, vh arrays
    """

    if not cols:
        df_: pd.DataFrame = df.copy()
    else:
        df_: pd.DataFrame = deepcopy(df.loc[:, cols])

    if normalize:
        for f in df_.keys():
            df_[f] = zero_mean_unit_variance(df_[f])

    u, s, vh = np.linalg.svd(df_, **kwargs)
    return SVD(u=u, s=s, vh=vh)


class OrderedSeries:
    def __init__(self, **data: Any):
        super().__init__(**data)
        raise TypeError(
            "Notice: OrderedSeries class has been supplanted by VectorSequence, see repackage() method"
        )



from typing import Tuple, Any, List, Dict

import matplotlib.pyplot as plt

from isthmuslib import VectorMultiset
from src.isthmuslib.logging import parse_string_with_embedded_json, parse_string_with_manual_tokens
from src.isthmuslib.utils import risky_cast, process_queue, recursive_batch_evaluation, return_best_input


def test_risky_cast():
    def val_and_type(input_value: Any) -> Tuple[Any, type]:
        return type(r := risky_cast(input_value)), r

    assert (int, 5) == val_and_type("5")
    assert (float, 4.55) == val_and_type("4.55")
    assert (str, "foo") == val_and_type("foo")
    assert (dict, {"a": 5, "b": 6}) == val_and_type("a=5, b=6")
    assert (bool, False) == val_and_type("False")
    assert (bool, True) == val_and_type("True")
    assert (dict, {"a": 5, "b": True}) == val_and_type("a=5, b=True")


def test_extract_json():
    s: str = """The quick
    [[EMBEDDED_JSON_LINE]]{"color": "red", "value": "#f00"}
    Brown fox jumped 
    [[EMBEDDED_JSON_LINE]]{"color": "green", "value": "#0f0"}
    [[EMBEDDED_JSON_LINE]]{"color": "blue", "value": "#00f"}
    Over the foobar
    [[EMBEDDED_JSON_LINE]]{"color": "cyan", "value": "#0ff"}
    [[EMBEDDED_JSON_LINE]]{"color": "black", "value": "#000"}
    """
    extracted = parse_string_with_embedded_json(s)
    assert extracted.iloc[1, 0] == "green"
    assert extracted.iloc[1, 1] == "#0f0"


def foobar(x: int) -> int:
    return x * 10


def starbar(x: int, y: int) -> str:
    return f"x={x} and y={y} so {x*y=}"


def test_process_queue():
    inputs: List[int] = list(range(25))
    star_inputs: List[Tuple[int, int]] = [(x, x + 2) for x in list(range(10))]

    result = process_queue(foobar, inputs, num_workers=5, cuts_per_worker=1, batching=False)
    assert result[:3] == [0, 10, 20]
    result = process_queue(foobar, inputs, num_workers=5, cuts_per_worker=1, batching=True)
    assert result[:3] == [0, 10, 20]
    result = process_queue(starbar, star_inputs, num_workers=5, cuts_per_worker=1, batching=False)
    assert result[:3] == ["x=0 and y=2 so x*y=0", "x=1 and y=3 so x*y=3", "x=2 and y=4 so x*y=8"]
    try:
        print(process_queue(starbar, star_inputs, num_workers=5, cuts_per_worker=1, batching=True))
    except NotImplementedError as e:
        print(f"... not implemented yet")


def test_parsing():
    log: str = """
    log-2018-08-31-20-07-29:2018-08-31 20:07:24.748	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49156/1651224 (0.282567 sec, 353.898367 blocks/sec), 99.548294 MB [49255:o][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.040	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49256/1651224 (0.291497 sec, 343.056704 blocks/sec), 99.046715 MB [49355:][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.280	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49356/1651224 (0.239763 sec, 417.078532 blocks/sec), 98.918274 MB [49455:oo][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.644	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49456/1651224 (0.363321 sec, 275.238701 blocks/sec), 98.767830 MB [49555:o.][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.981	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49556/1651224 (0.335829 sec, 297.770592 blocks/sec), 98.502403 MB [49655:...][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:26.264	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49656/1651224 (0.281597 sec, 355.117420 blocks/sec), 98.384758 MB [49755:oo....][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:26.776	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49756/1651224 (0.404512 sec, 247.211455 blocks/sec), 98.054283 MB [49855:oo.....][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:27.066	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49856/1651224 (0.289512 sec, 345.408826 blocks/sec), 98.258347 MB [49955:oo.....][0m
    """

    # Set the rules for extracting the vectors
    record_delimiter: str = "lo"
    tokens_dictionary: Dict[str, Tuple[str, str]] = {
        # varname: (left_token, right_token)
        "date_time_stamp": ("g-", ":"),
        "height": ("Synced ", "/"),
        "time_to_load": ("(", " sec"),
    }

    # Extract the text into a timeseries
    timeseries: VectorMultiset = VectorMultiset(
        data=parse_string_with_manual_tokens(
            log,
            record_delimiter=record_delimiter,
            parallelize_processing=True,
            tokens_dictionary=tokens_dictionary,
        )
    )
    print(timeseries.data)


def mock_hill(kwargs) -> float:
    x = kwargs.get("x")
    y = kwargs.get("y")
    return 1 / x + 1 / (y**2)


def test_recursive_batch_evaluation():
    inital_point: Dict[str, Any] = {"x": 400, "y": 300}
    current_best_input, history = recursive_batch_evaluation(
        func=mock_hill,
        initial_input=inital_point,
        selection_method=max,
        batch_generator_kwargs={"width_prct": 10, "num_samples": 3},
        max_deep=5,
        print_progress=True,
        num_workers=64,
        return_history=True,
    )
    assert current_best_input["x"] < inital_point["x"]
    assert current_best_input["y"] < inital_point["y"]


def test_selector():
    inputs: List[Any] = ["abc", "def", "efg"]
    eval_outputs: List[Dict[str, Any]] = [
        {"foobar": "xyz", "fitness": 4},
        {"fitness": 99, "baz": 12345},
        {"fitness": -5},
    ]
    best = return_best_input(inputs, eval_outputs)
    assert best == "def"


def y(x: float) -> float:
    return 10 - (x - 5) ** 2


def y_return_scalar(d) -> float:
    return y(d.get("x"))


def test_recursive_batch_evaluation_new_with_scalar(show_plot: bool = True):
    # Set up the run
    start_at: float = 15

    best_x, run_data = recursive_batch_evaluation(
        func=y_return_scalar,
        initial_input={"x": start_at},
        max_deep=5,
        return_history=True,
        evaluate_initial_inputs=True,
        infinite_memory=True,
        selection_method=max,
        catch_exceptions=False,
        batch_generator_kwargs={
            "width_prct": 50,
            "num_samples": 5,
            "fields": ["x"],
            "width_temperature_prct": 0,
        },
    )

    print(f"{best_x=}\n{run_data}")
    if show_plot:
        x_vals: List[float] = list(range(-10, 25))
        plt.figure(facecolor="w")
        plt.plot(x_vals, [y_return_scalar({"x": x}) for x in x_vals], color="k")
        x_vec = [d.get("in").get("x") for d in run_data]
        y_vec = [x.get("out") for x in run_data]
        plt.scatter(x_vec, y_vec, c=list(range(len(x_vec))), s=50)
        plt.show()

    assert best_x["x"] < start_at


def y_return_dict(d) -> dict:
    return {"y": y(d.get("x"))}


def test_recursive_batch_evaluation_new_with_dict(show_plot: bool = True):
    # Set up the run
    start_at: float = 15

    best_x, run_data = recursive_batch_evaluation(
        func=y_return_dict,
        initial_input={"x": start_at},
        max_deep=5,
        return_history=True,
        evaluate_initial_inputs=True,
        infinite_memory=True,
        selection_method="y",
        catch_exceptions=False,
        batch_generator_kwargs={
            "width_prct": 50,
            "num_samples": 5,
            "fields": ["x"],
            "width_temperature_prct": 0,
        },
    )

    print(f"{best_x=}\n{run_data}")
    if show_plot:
        x_vals: List[float] = list(range(-10, 25))
        plt.figure(facecolor="w")
        plt.plot(x_vals, [y_return_dict({"x": x})["y"] for x in x_vals], color="k")
        x_vec = [d.get("in").get("x") for d in run_data]
        y_vec = [x.get("out").get("y") for x in run_data]
        plt.scatter(x_vec, y_vec, c=list(range(len(x_vec))), s=50)
        plt.show()

    print(f"{run_data[-1]['in']}")
    assert run_data[-1]["in"]["incremented_counter"] == 4
    # assert run_data[-1]['in']['clock_time_at_start'] > 100000
    assert best_x["x"] < start_at



from pydantic import BaseModel
from typing import Tuple, Any, Dict, Optional
from .utils import Rosetta
from copy import deepcopy


class Style(BaseModel):
    """Miscellaneous properties; set defaults once, then use or inherit elsewhere for consistent style
    (note: some names are formatted to be congruent with matplotlib.pyplot vars & args, rather than isthmuslib)"""

    color: Optional[Any] = "DarkGreen"  # isthmuslib default
    facecolor: Optional[Any] = "w"
    title_fontsize: Optional[Any] = 18.0
    label_fontsize: Optional[Any] = 15.0
    legend_fontsize: Optional[Any] = 15.0
    tick_fontsize: Optional[Any] = 12.0
    figsize: Optional[Tuple[Any, Any]] = (12.0, 8.0)
    linewidth: Optional[Any] = 5.0
    linestyle: Optional[Any] = "-"
    grid: Optional[bool] = False
    tight_axes: Optional[bool] = True
    formatter: Optional[str] = "%Y-%m-%d %H:%M:%S"
    include_timezone: Optional[bool] = False
    markersize: Optional[Any] = 50
    transparent_alpha: Optional[float] = 0.5
    watermark_placement: Optional[Any] = (0.05, 0.05)
    watermark_fontsize: Optional[Any] = 20
    watermark_color: Optional[Any] = "dimgrey"
    watermark_text: Optional[str] = ""
    histogram_bins: Optional[Any] = 25
    multi_hist_alpha: Optional[float] = 0.5
    rosetta: Optional[Rosetta] = Rosetta()
    median_linestyle: Optional[Any] = "--"
    median_linewidth: Optional[Any] = 3.0
    median_linecolor: Optional[Any] = "k"
    mean_linestyle: Optional[Any] = "-"
    mean_linewidth: Optional[Any] = 3.0
    mean_linecolor: Optional[Any] = "k"
    cycler: Optional[Any] = None
    good_color: Optional[Any] = "navy"
    bad_color: Optional[Any] = "firebrick"
    cmap: Optional[Any] = "inferno"
    sequential_cmap: Optional[Any] = "Greens"
    log_formatter: Optional[
        str
    ] = "\n@@ {time:x} AT: {time} | LEVEL: {level} | IN: {name}.{function}\n\n{message} |\n"
    timeframe_prefix: Optional[str] = "from "
    timeframe_between: Optional[str] = " to "
    timeframe_suffix: Optional[str] = ""
    auto_locf_params: Optional[Dict[str, Any]] = {
        "window": 3,
        "impute_method": "median",
        "impute_direction": "forward",
        "add_noise": False,
    }
    x_axis_human_tick_labels: Optional[bool] = False
    x_axis_formatter: Optional[str] = None
    dict_pretty_max_length: Optional[int] = 32
    dict_pretty_previews: Optional[bool] = False
    axhline_color: Optional[str] = "gray"
    axhline_linestyle: Optional[str] = "--"
    axhline_linewidth: Optional[float] = 1.0
    axvline_color: Optional[str] = "gray"
    axvline_linestyle: Optional[str] = "--"
    axvline_linewidth: Optional[float] = 1.0

    def translate(self, key: str, **kwargs) -> str:
        """Helper function that allows Style objects to translate text according to the provided Rosetta mappings

        :param key: the string to be translated
        :param kwargs: keyword arguments for the rosetta translate() method
        :return: translated string
        """
        if self.rosetta:
            return self.rosetta.translate(key, **kwargs)
        else:
            return key

    def override(self, *args):
        """Helper function that overrides specific style features

        :return: returns a fresh copy of a Style object with the updated values
        """

        if not args:
            return deepcopy(self)

        elif len(args) == 1:
            if args[0] is None:
                return deepcopy(self)
            elif isinstance(args[0], dict):
                override_dict: Dict[str, any] = args[0]
            elif getattr(args[0], "dict", None):
                override_dict: Dict[str, any] = args[0].dict()
            else:
                raise ValueError(f"Unsure how to interpret override input of type {type(args[0])}")

        elif (len(args) == 2) and isinstance(args[0], str):
            override_dict: Dict[str, any] = {args[0]: args[1]}

        else:
            raise ValueError(f"Unsure how to interpret 3+ inputs to override")

        return self.__class__(**{**self.dict(), **override_dict})
